from __future__ import division

import numpy as np
import statsmodels.sandbox.distributions.mv_normal as mvd
from scipy import linalg, stats
from scipy.special import multigammaln
from sklearn.base import BaseEstimator
from sklearn.datasets import make_sparse_spd_matrix
from sklearn.utils.extmath import fast_logdet


class WishartDistribution(BaseEstimator):
    """Wishart distribution."""

    def __init__(self, nu, S):
        self.nu = nu
        self.S = S
        self.D = S.shape[0]
        self.inv_S = linalg.pinvh(S)

    @property
    def mean(self):
        return self.nu * self.S

    @property
    def mode(self):
        return (self.nu - self.D - 1) * self.S \
            if self.nu > self.D + 1 else np.nan

    def log_likelihood(self, X):
        """Equivalent to scipy.

        from scipy.stats import wishart
        wishart.logpdf(X, nu, S)
        """
        nu = self.nu
        n_dim = X.shape[0]
        inv_S = self.inv_S

        logp = (nu - n_dim - 1) * fast_logdet(X)
        logp -= np.sum(X * inv_S)
        logp -= nu * n_dim * np.log(2)
        logp -= 2 * multigammaln(0.5 * nu, n_dim)
        logp -= nu * fast_logdet(self.S)
        logp /= 2.
        return logp

    def likelihood(self, X):
        return np.exp(self.log_likelihood(X))

    def sample(self, n_samples=1):
        # return distributions.rwishart(self.nu, self.psi)
        # return distributions.rwishart(self.nu, self.inv_psi)
        return stats.wishart(df=self.nu, scale=self.S, size=n_samples)


class InverseWishartDistribution(WishartDistribution):
    """Inverse Wishart (IW) distribution."""

    def __init__(self, nu, S):
        super(InverseWishartDistribution, self).__init__(nu=nu, S=S)

    @property
    def mean(self):
        # return self.inv_S / (self.nu - self.D - 1)
        return self.S / (self.nu - self.D - 1)

    @property
    def mode(self):
        # return self.inv_S / (self.nu + self.D + 1)
        return self.S / (self.nu + self.D + 1)

    def log_likelihood(self, X):
        """Equivalent to scipy.

        from scipy.stats import invwishart
        invwishart.logpdf(X, nu, S)
        """
        nu = self.nu
        n_dim = X.shape[0]

        logp = nu * fast_logdet(self.S)
        logp -= np.sum(self.S * linalg.pinvh(X))
        logp -= (nu + n_dim + 1) * fast_logdet(X)
        logp -= nu * n_dim * np.log(2)
        logp -= 2 * multigammaln(0.5 * nu, n_dim)
        logp /= 2.
        return logp

    #def wishartrand(self):
    #    dim = self.inv_psi.shape[0]
    #    chol = np.linalg.cholesky(self.inv_psi)
    #    foo = np.zeros((dim,dim))
    #
    #    for i in range(dim):
    #        for j in range(i+1):
    #            if i == j:
    #                foo[i,j] = np.sqrt(chi2.rvs(self.nu-(i+1)+1))
    #            else:
    #                foo[i,j]  = np.random.normal(0,1)
    #    return np.dot(chol, np.dot(foo, np.dot(foo.T, chol.T)))
    def sample(self, n_samples=1):
        # return distributions.rwishart(self.nu, self.psi)
        # return distributions.rwishart(self.nu, self.inv_psi)
        return stats.invwishart.rvs(df=self.nu, scale=self.S, size=n_samples)


class NormalInverseWishartDistribution(InverseWishartDistribution):
    """Normal-inverse-wishart (NIW) distribution.

    Defined as Normal x InverseWishart distribution.

    References
    ----------
    Murphy's "A probabilistic approach", pag 133.
    """

    def __init__(self, mu, kappa, nu, S):
        super(NormalInverseWishartDistribution, self).__init__(nu=nu, S=S)
        self.mu = mu
        self.kappa = kappa

    def sample(self, n_samples=1):
        """A sample of the NIW is a pair (mu, Sigma).

        Sigma is a sample from the InverseWishartDistribution.
        mu is a sample from the multivariate_normal.
        """
        iw_samples = super(NormalInverseWishartDistribution, self).sample(
            n_samples=n_samples)
        if n_samples == 1:
            mu = np.random.multivariate_normal(self.mu, iw_samples / self.kappa)
            return (mu, iw_samples)
        return [(np.random.multivariate_normal(self.mu, Sigma / self.kappa),
                 Sigma) for Sigma in iw_samples]

    def log_likelihood(self, mu, X):
        """Equivalent to likelihood of Normal times InverseWishart."""
        logl = super(NormalInverseWishartDistribution, self).log_likelihood(X)
        logl += stats.multivariate_normal.logpdf(mu, self.mu, X / self.kappa)

        # # equivalent to
        # nu = self.nu
        # n_dim = X.shape[0]
        #
        # logp = nu * fast_logdet(self.S)
        # logp -= np.sum(self.S * linalg.pinvh(X))
        # logp -= (nu + n_dim + 2) * fast_logdet(X)
        # logp -= nu * n_dim * np.log(2)
        # logp -= 2 * multigammaln(0.5 * nu, n_dim)
        #
        # logp -= n_dim * np.log(2 * np.pi / self.kappa)
        # logp -= self.kappa * np.linalg.multi_dot(
        #     (mu - self.mu, linalg.pinvh(X), mu - self.mu))
        # logp /= 2.

        return logl

    @property
    def mode(self):
        """Mode of the distribution (Murphy, pag 134)."""
        return (self.mu, self.S / (self.nu + self.D + 2))

    @property
    def marginals(self):
        """Marginals of the distribution (Murphy, pag 134)."""
        # mu has a multivariate student t distribution
        df = self.nu - self.D + 1
        # XXX this is univariate, need multivariate in Python
        # marginal_mu = stats.t(
        #     df=df, loc=self.mu, scale=self.S / (self.kappa * df))
        marginal_mu = mvd.MVT(
            df=df, mean=self.mu.ravel(),
            sigma=self.S / (self.kappa * df))
        marginal_sigma = InverseWishartDistribution(nu=self.nu, S=self.S)
        return (marginal_mu, marginal_sigma)

    def posterior(self, X):
        """The posterior can be shown to be NIW with updated parameters.

        References
        ----------
        Probabilistic Approach, Murphy (pag 134).
        """
        n = X.shape[0]
        nu_n = self.nu + n
        S = X.T.dot(X)

        mean = np.mean(X, axis=0)
        # S_mu = empirical_covariance(data - mean, assume_centered=True) * n

        k_0 = self.kappa
        k_n = k_0 + n

        m_0 = self.mu
        m_n = (k_0 * m_0 + n * mean) / k_n

        m_0 = np.atleast_2d(m_0)
        m_n = np.atleast_2d(m_n)

        # mm = np.atleast_2d(mean - m_0)
        # psi_n_v1 = self.psi + S_mu + k_0 * n / k_n * mm.T.dot(mm)
        S_n = self.S + S + k_0 * m_0.T.dot(m_0) - k_n * m_n.T.dot(m_n)

        return NormalInverseWishartDistribution(m_n, k_n, nu_n, S_n)

    def predictive(self):
        """Posterior predictive. See pag 135."""
        df = self.nu - self.D + 1
        return mvd.MVT(
            df=df, mean=self.mu.ravel(),
            sigma=(self.kappa + 1) * self.S / (self.kappa * df))


def main():
    n_dim = 5
    n_samples = 100
    v_0 = n_dim + 2 + 1  # v_0 > p + 2
    Sigma = make_sparse_spd_matrix(n_dim, 0.75)
    mu = np.zeros(n_dim) - 3

    x = NormalInverseWishartDistribution(mu, 1, v_0, Sigma)
    # X = x.sample(n_samples=n_samples)
    X = np.random.multivariate_normal(mu, Sigma, size=n_samples)

    # prior
    prior = NormalInverseWishartDistribution(
        np.zeros(n_dim), 1, v_0, np.eye(n_dim))

    z = prior.posterior(X)

    print x
    print z
    print np.linalg.norm(z.mode[1] - Sigma)
    print np.linalg.norm(z.marginal_mode - Sigma)
    print np.linalg.norm(z.marginal_mean - Sigma)


if __name__ == '__main__':
    main()



from __future__ import division, print_function

import warnings

import numpy as np
import six
from scipy import sparse
from six.moves import range
from sklearn.base import RegressorMixin
from sklearn.linear_model.base import (LinearClassifierMixin, LinearModel,
                                       _pre_fit)
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils import check_array, check_X_y, deprecated
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.utils.validation import check_is_fitted

from regain.prox import soft_thresholding
from regain.utils import flatten
from regain.wrapper.paspal.glopridu import glopridu_algorithm

try:
    from regain.wrapper.paspal.wrapper import group_lasso_overlap_paspal
    _MATLAB_FOUND_ = True
except ImportError:
    _MATLAB_FOUND_ = False


def _remove_unused_features(data, groups):
    unique_idx = sorted(set(flatten(groups)))
    hashing = dict(zip(unique_idx, range(len(unique_idx))))

    new_groups = [[hashing[g] for g in group] for group in groups]
    return data[:, unique_idx], new_groups


def D_function(d, groups):
    D = np.zeros(d)
    for dimension in range(d):
        # find groups which contain this dimension
        for g, group in enumerate(groups):
            if dimension in group:
                D[dimension] += 1
    return D


def P_star_x_bar_function(x, d, groups):
    P_star_x_bar = np.zeros(d)
    for dim in range(d):
        ss = 0
        count = 0
        for g, group in enumerate(groups):
            if dim in group:
                idx = np.argwhere(np.array(group) == dim)[0]
                ss += x[g][idx]
                count += 1
        if count > 0:
            ss /= count
        P_star_x_bar[dim] = ss
    return P_star_x_bar


def group_lasso_overlap(
        A, b, lamda=1.0, groups=None, rho=1.0, max_iter=100, tol=1e-4,
        verbose=False, rtol=1e-2):
    n, d = A.shape

    x = [np.zeros(len(g)) for g in groups]  # local variables
    z = np.zeros(d)
    y = [np.zeros(len(g)) for g in groups]

    D = np.diag(D_function(d, groups))
    Atb = A.T.dot(b)
    inv = np.linalg.inv(A.T.dot(A) + rho * D)
    hist = []
    count = 0
    for k in range(max_iter):
        # x update
        for i, g in enumerate(groups):
            x[i] = soft_thresholding(x[i] - y[i] / rho, lamda / rho)

        # z update
        zold = z
        x_consensus = P_star_x_bar_function(x, d, groups)
        y_consensus = P_star_x_bar_function(y, d, groups)
        z = inv.dot(Atb + D.dot(y_consensus + rho * x_consensus))

        for i, g in enumerate(groups):
            y[i] += rho * (x[i] - z[g])

        # diagnostics, reporting, termination checks
        history = (
            objective(A, b, lamda, x, z),  # objective
            np.linalg.norm(x_consensus - z),  # rnorm
            np.linalg.norm(-rho * (z - zold)),  # snorm
            np.sqrt(d) * tol + rtol * max(
                np.linalg.norm(x_consensus), np.linalg.norm(-z)),  # eps primal
            np.sqrt(d) * tol + rtol * np.linalg.norm(rho * y_consensus)
            # eps dual
        )

        if verbose:
            print(
                "obj: %.4f, rnorm: %.4f, snorm: %.4f,"
                "eps_pri: %.4f, eps_dual: %.4f" % history)

        hist.append(history)
        if history[1] < history[3] and history[2] < history[4]:
            if count > 10:
                break
            else:
                count += 1
        else:
            count = 0

    return z, hist, k


class GroupLassoOverlap(LinearModel, RegressorMixin):
    def __init__(
            self, alpha=1.0, fit_intercept=True, groups=None, rho=1.0,
            n_jobs=1, tol=1e-4, verbose=False, rtol=1e-2, normalize=False,
            precompute=False, max_iter=1000, copy_X=True, warm_start=False,
            positive=False, random_state=None, selection='cyclic', mode='admm',
            matlab_engine=None):
        self.alpha = alpha
        self.coef_ = None
        self.fit_intercept = fit_intercept
        self.groups = groups
        self.rho = rho
        self.verbose = verbose
        self.normalize = normalize
        self.precompute = precompute
        self.max_iter = max_iter
        self.copy_X = copy_X
        self.tol = tol
        self.rtol = rtol
        self.warm_start = warm_start
        self.positive = positive
        self.intercept_ = 0.0
        self.random_state = random_state
        self.selection = selection
        self.matlab_engine = matlab_engine
        self.n_jobs = n_jobs

        self.mode = mode
        if mode == 'paspal-matlab' and not _MATLAB_FOUND_:
            raise ValueError(
                "Cannot use Matlab implementation. Use `mode='admm'` or `mode='paspal'`."
            )

    def fit(self, X, y, check_input=True):
        """Fit model with coordinate descent.

        Parameters
        -----------
        X : ndarray or scipy.sparse matrix, (n_samples, n_features)
            Data

        y : ndarray, shape (n_samples,) or (n_samples, n_targets)
            Target

        check_input : boolean, (default=True)
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        Notes
        -----

        Coordinate descent is an algorithm that considers each column of
        data at a time hence it will automatically convert the X input
        as a Fortran-contiguous numpy array if necessary.

        To avoid memory re-allocation it is advised to allocate the
        initial data in memory directly using that format.
        """

        if self.alpha == 0:
            warnings.warn(
                "With alpha=0, this algorithm does not converge "
                "well. You are advised to use the LinearRegression "
                "estimator", stacklevel=2)

        if isinstance(self.precompute, six.string_types):
            raise ValueError(
                'precompute should be one of True, False or'
                ' array-like. Got %r' % self.precompute)

        # We expect X and y to be float64 or float32 Fortran ordered arrays
        # when bypassing checks
        if check_input:
            X, y = check_X_y(
                X, y, accept_sparse='csc', order='F',
                dtype=[np.float64, np.float32], copy=self.copy_X
                and self.fit_intercept, multi_output=True, y_numeric=True)
            y = check_array(
                y, order='F', copy=False, dtype=X.dtype.type, ensure_2d=False)

        X, y, X_offset, y_offset, X_scale, precompute, Xy = \
            _pre_fit(X, y, None, self.precompute, self.normalize,
                     self.fit_intercept, copy=False)

        if y.ndim == 1:
            y = y[:, None]
        if Xy is not None and Xy.ndim == 1:
            Xy = Xy[:, None]

        n_samples, n_features = X.shape
        n_targets = y.shape[1]

        if self.selection not in ['cyclic', 'random']:
            raise ValueError("selection should be either random or cyclic.")

        if not self.warm_start or self.coef_ is None:
            coef_ = np.zeros((n_targets, n_features), dtype=X.dtype, order='F')
        else:
            coef_ = self.coef_
            if coef_.ndim == 1:
                coef_ = coef_[None, :]

        dual_gaps_ = np.zeros(n_targets, dtype=X.dtype)
        if self.n_jobs == 1:
            self.n_iter_ = []
            history = []
            for k in range(n_targets):
                if self.mode == 'admm':
                    this_coef, hist, this_iter = \
                        group_lasso_overlap(
                            X, y[:, k], lamda=self.alpha, groups=self.groups,
                            rho=self.rho, max_iter=self.max_iter, tol=self.tol,
                            verbose=self.verbose, rtol=self.rtol)
                elif self.mode == 'paspal-matlab':
                    this_coef, hist, this_iter = \
                        group_lasso_overlap_paspal(
                            X, y[:, k], lamda=self.alpha, groups=self.groups,
                            rho=self.rho, max_iter=self.max_iter, tol=self.tol,
                            verbose=self.verbose, rtol=self.rtol,
                            matlab_engine=self.matlab_engine)
                elif self.mode == 'paspal':  # paspal wrapper
                    this_coef, hist, this_iter = \
                        glopridu_algorithm(
                            X, y[:, k], tau=self.alpha, blocks=self.groups,
                            max_iter_ext=self.max_iter, tol_ext=self.tol,
                            verbose=self.verbose, tol_int=self.rtol)
                else:
                    raise ValueError(self.mode)
                coef_[k] = this_coef.ravel()
                history.append(hist)
                self.n_iter_.append(this_iter)
        else:
            import joblib as jl
            if self.mode == 'admm':
                coef_, history, self.n_iter_ = \
                    zip(*jl.Parallel(n_jobs=self.n_jobs)(
                        jl.delayed(group_lasso_overlap)(
                            X, y[:, k], lamda=self.alpha, groups=self.groups,
                            rho=self.rho, max_iter=self.max_iter, tol=self.tol,
                            verbose=self.verbose, rtol=self.rtol)
                        for k in range(n_targets)))
            elif self.mode == 'paspal-matlab':  # paspal wrapper
                coef_, history, self.n_iter_ = \
                    zip(*jl.Parallel(n_jobs=self.n_jobs)(
                        jl.delayed(group_lasso_overlap_paspal)(
                            X, y[:, k], lamda=self.alpha, groups=self.groups,
                            rho=self.rho, max_iter=self.max_iter, tol=self.tol,
                            verbose=self.verbose, rtol=self.rtol,
                            matlab_engine=self.matlab_engine)
                        for k in range(n_targets)))
            elif self.mode == 'paspal':  # paspal wrapper
                coef_, history, self.n_iter_ = \
                    zip(*jl.Parallel(n_jobs=self.n_jobs)(
                        jl.delayed(glopridu_algorithm)(
                            X, y[:, k], tau=self.alpha, blocks=self.groups,
                            max_iter_ext=self.max_iter, tol_ext=self.tol,
                            verbose=self.verbose, tol_int=self.rtol)
                        for k in range(n_targets)))
            else:
                raise ValueError(self.mode)

        if n_targets == 1:
            self.n_iter_ = self.n_iter_[0]

        self.coef_, self.dual_gap_ = map(np.squeeze, [coef_, dual_gaps_])
        self._set_intercept(X_offset, y_offset, X_scale)

        # workaround since _set_intercept will cast self.coef_ into float64
        self.coef_ = np.asarray(self.coef_, dtype=X.dtype)

        self.history_ = history

        # return self for chaining fit and predict calls
        return self

    @property
    def sparse_coef_(self):
        """ sparse representation of the fitted ``coef_`` """
        return sparse.csr_matrix(self.coef_)

    @deprecated(" and will be removed in 0.19")
    def decision_function(self, X):
        """Decision function of the linear model

        Parameters
        ----------
        X : numpy array or scipy.sparse matrix of shape (n_samples, n_features)

        Returns
        -------
        T : array, shape (n_samples,)
            The predicted decision function
        """
        return self._decision_function(X)

    def _decision_function(self, X):
        """Decision function of the linear model

        Parameters
        ----------
        X : numpy array or scipy.sparse matrix of shape (n_samples, n_features)

        Returns
        -------
        T : array, shape (n_samples,)
            The predicted decision function
        """
        check_is_fitted(self, 'n_iter_')
        if sparse.isspmatrix(X):
            return safe_sparse_dot(
                X, self.coef_.T, dense_output=True) + self.intercept_
        else:
            return super(GroupLassoOverlap, self)._decision_function(X)


class GroupLassoOverlapClassifier(LinearClassifierMixin, GroupLassoOverlap):
    """Class to extend group lasso in case of classification."""

    def fit(self, X, y, check_input=True):
        self._label_binarizer = LabelBinarizer(pos_label=1, neg_label=-1)
        Y = self._label_binarizer.fit_transform(y)
        if self._label_binarizer.y_type_.startswith('multilabel'):
            # we don't (yet) support multi-label classification in ENet
            raise ValueError(
                "%s doesn't support multi-label classification" %
                (self.__class__.__name__))

        # Y = column_or_1d(Y, warn=True)
        super(GroupLassoOverlapClassifier, self).fit(X, Y)
        if self.classes_.shape[0] > 2:
            ndim = self.classes_.shape[0]
        else:
            ndim = 1
        self.coef_ = self.coef_.reshape(ndim, -1)

        return self

    @property
    def classes_(self):
        return self._label_binarizer.classes_


def _overlapping_group_lasso(
        A, b, lamda=1.0, groups=None, rho=1.0, alpha=1.0, max_iter=100,
        tol=1e-4):
    # % solves the following problem via ADMM:
    # %   minimize 1/2*|| Ax - b ||_2^2 + \lambda sum(norm(x_i))
    # %
    # % The input p is a K-element vector giving the block sizes n_i, so that x_i
    # % is in R^{n_i}.
    # % The solution is returned in the vector x.
    # %
    # % history is a structure that contains the objective value, the primal and
    # % dual residual norms, and the tolerances for the primal and dual residual
    # % norms at each iteration.
    # % rho is the augmented Lagrangian parameter.
    # % alpha is the over-relaxation parameter (typical values for alpha are
    # % between 1.0 and 1.8).
    rtol = 1e-2

    n, d = A.shape
    N = len(groups)

    x = [np.zeros(len(g)) for g in groups]  # local variables
    z = np.zeros(d)
    y = np.zeros(d)

    hist = []
    AtA = A.T.dot(A)
    Atb = A.T.dot(b)
    inverse = np.linalg.inv(N * AtA + rho * np.eye(d))
    for k in range(max_iter):
        # % x-update (to be done in parallel)
        P_star_xk_bar = P_star_x_bar_function(x)
        for i, g in enumerate(groups):
            # x update; update each local x
            x[i] = prox(x[i] + (z - P_star_xk_bar - y / rho)[g], lamda / rho)

        P_star_xk1_bar = P_star_x_bar_function(x)
        z = inverse.dot(Atb + rho * P_star_xk1_bar + y)

        # y-update
        y += rho * (P_star_xk1_bar - z)

        # # % compute the dual residual norm square
        # s = 0
        # q = 0
        # zsold = zs
        # zs = z.reshape(-1,1).dot(np.ones((1, N)))
        # zs += Aixi - Axbar.reshape(-1,1).dot(np.ones((1, N)))
        # for i in range(N):
        #     # % dual residual norm square
        #     s = s + np.linalg.norm(-rho * Ats[i].dot((zs[:,i] - zsold[:,i])))**2
        #     # % dual residual epsilon
        #     q = q + np.linalg.norm(rho*Ats[i].dot(u))**2;
        #
        # # % diagnostics, reporting, termination checks
        # history = []
        # history.append(objective(A, b, lamda, N, x, z))
        # history.append(np.sqrt(N)*np.linalg.norm(z - Axbar))
        # history.append(np.sqrt(s))
        #
        # history.append(np.sqrt(n)*ABSTOL + rtol*max(np.linalg.norm(Aixi,'fro'), np.linalg.norm(-zs, 'fro')))
        # history.append(np.sqrt(n)*ABSTOL + rtol*np.sqrt(q))
        #
        # hist.append(history)
        # if history[1] < history[3] and history[2] < history[4]:
        #     break

    return P_star_x_bar_function(x), x


def fi(xi):
    return np.linalg.norm(xi)


def objective(A, b, alpha, x, z):
    return .5 * np.sum((A.dot(z) - b)**2) + alpha * np.sum(
        [np.linalg.norm(x_i) for x_i in x])



"""Graphical latent variable model selection via ADMM."""
from __future__ import division

import warnings

import numpy as np
from scipy import linalg
from six.moves import range
from sklearn.utils.extmath import squared_norm

from regain.norm import l1_od_norm
from regain.prox import prox_laplacian, prox_trace_indicator, soft_thresholding
from regain.update_rules import update_rho
from regain.utils import convergence


def objective(S, R, K, L, alpha, tau):
    """Objective function for latent graphical lasso."""
    obj = 0.5 * squared_norm(S - R)
    obj += alpha * l1_od_norm(K)
    obj += tau * np.linalg.norm(L, ord='nuc')
    return obj


def infimal_convolution(
        S, alpha=1., tau=1., rho=1., max_iter=100,
        verbose=False, tol=1e-4, rtol=1e-2, return_history=False,
        return_n_iter=True, update_rho_options=None, compute_objective=True):
    r"""Latent variable graphical lasso solver.

    Solves the following problem via ADMM:
        min - log_likelihood(S, K-L) + alpha ||K||_{od,1} + tau ||L_i||_*

    where S is the empirical covariance of the data
    matrix D (training observations by features).

    Parameters
    ----------
    emp_cov : array-like
        Empirical covariance matrix.
    alpha, tau : float, optional
        Regularisation parameters.
    rho : float, optional
        Augmented Lagrangian parameter.
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.
    return_n_iter : bool, optional
        Return the number of iteration before convergence.
    verbose : bool, default False
        Print info at each iteration.

    Returns
    -------
    K, L : np.array, 2-dimensional, size (d x d)
        Solution to the problem.
    S : np.array, 2 dimensional
        Empirical covariance matrix.
    n_iter : int
        If return_n_iter, returns the number of iterations before convergence.
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    K = np.zeros_like(S)
    L = np.zeros_like(S)
    U = np.zeros_like(S)
    R_old = np.zeros_like(S)

    checks = []
    for iteration_ in range(max_iter):
        # update R
        A = K - L - U
        A += A.T
        A /= 2.
        R = prox_laplacian(S + rho * A, lamda=rho / 2.)

        A = L + R + U
        K = soft_thresholding(A, lamda=alpha / rho)

        A = K - R - U
        A += A.T
        A /= 2.
        L = prox_trace_indicator(A, lamda=tau / rho)

        # update residuals
        U += R - K + L

        # diagnostics, reporting, termination checks
        obj = objective(S, R, K, L, alpha, tau) \
            if compute_objective else np.nan
        rnorm = np.linalg.norm(R - K + L)
        snorm = rho * np.linalg.norm(R - R_old)
        check = convergence(
            obj=obj, rnorm=rnorm, snorm=snorm,
            e_pri=np.sqrt(R.size) * tol + rtol * max(
                np.linalg.norm(R), np.linalg.norm(K - L)),
            e_dual=np.sqrt(R.size) * tol + rtol * rho * np.linalg.norm(U)
        )
        R_old = R.copy()

        if verbose:
            print("obj: %.4f, rnorm: %.4f, snorm: %.4f,"
                  "eps_pri: %.4f, eps_dual: %.4f" % check[:5])

        checks.append(check)
        if check.rnorm <= check.e_pri and check.snorm <= check.e_dual:
            break
        if check.obj == np.inf:
            break
        rho_new = update_rho(rho, rnorm, snorm, iteration=iteration_,
                             **(update_rho_options or {}))
        # scaled dual variables should be also rescaled
        U *= rho / rho_new
        rho = rho_new
    else:
        warnings.warn("Objective did not converge.")

    covariance_ = linalg.pinvh(K)
    return_list = [K, L, covariance_]
    if return_history:
        return_list.append(checks)
    if return_n_iter:
        return_list.append(iteration_)
    return return_list



import matplotlib.pyplot as plt
import networkx as nx
import numpy as np


def plot_graph_with_latent_variables(
        adjacency_matrix, n_observed, title="", n_latents=0, labels=None,
        ax=None, node_size=100, font_size=5):
    """Plot graph with latent variables."""
    import networkx as nx
    # plt.figure(figsize=(15,10))
    if ax is None:
        f, ax = plt.subplots(1, 1)

    G = nx.from_numpy_matrix(adjacency_matrix)
    pos = nx.circular_layout(G, scale=.1)

    if n_latents > 0:
        nx.draw_networkx_nodes(
            G, pos, nodelist=list(np.arange(0, n_latents)), ax=ax,
            node_color='r', node_size=node_size, alpha=0.8)

    if labels is not None:
        nx.draw_networkx_labels(G, pos, labels, font_size=font_size, ax=ax)

    nx.draw(
        G, pos, node_color='#A0CBE2', node_size=node_size, edge_color=range(
            len(G.edges.values())), alpha=1, width=2, edge_cmap=plt.cm.Blues,
        with_labels=False, ax=ax)

    ax.set_axis_off()
    ax.set_title(title)
    return ax



"""Sparse inverse covariance selection via ADMM.

More information can be found in the paper linked at:
http://www.stanford.edu/~boyd/papers/distr_opt_stat_learning_admm.html
"""
from __future__ import division

import warnings

import numpy as np
from six.moves import range
from sklearn.covariance import empirical_covariance
from sklearn.utils.extmath import fast_logdet
from sklearn.utils.validation import check_array

from regain.norm import l1_od_norm
from regain.prox import prox_logdet, soft_thresholding_od
from regain.update_rules import update_rho
from regain.utils import convergence

try:
    # sklean >= 0.20
    from sklearn.covariance import GraphicalLasso as GraphLasso
except ImportError:
    # sklearn < 0.20
    from sklearn.covariance import GraphLasso


def logl(emp_cov, precision):
    """Gaussian log-likelihood without constant term."""
    return fast_logdet(precision) - np.sum(emp_cov * precision)


def objective(emp_cov, x, z, alpha):
    """Graph lasso objective."""
    return -logl(emp_cov, x) + alpha * l1_od_norm(z)


def graph_lasso(
        emp_cov, alpha=0.01, rho=1, over_relax=1, max_iter=100, verbose=False,
        tol=1e-4, rtol=1e-4, return_history=False, return_n_iter=True,
        update_rho_options=None, compute_objective=True, mode='admm'):
    """Graph lasso solver via ADMM.

    Solves the following problem via ADMM:
        minimize  trace(S*X) - log det X + alpha ||X||_{od,1}

    where S is the empirical covariance of the data
    matrix D (training observations by features).

    Parameters
    ----------
    emp_cov : array-like
        Empirical covariance matrix.
    alpha : float, optional
        Regularisation parameter.
    rho : float, optional
        Augmented Lagrangian parameter.
    over_relax : float, optional
        Over-relaxation parameter (typically between 1.0 and 1.8).
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.
    return_n_iter : bool, optional
        Return the number of iteration before convergence.
    verbose : bool, default False
        Print info at each iteration.

    Returns
    -------
    X : numpy.array, 2-dimensional
        Solution to the problem.
    S : np.array, 2 dimensional
        Empirical covariance matrix.
    n_iter : int
        If return_n_iter, returns the number of iterations before convergence.
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    _, n_features = emp_cov.shape
    covariance_ = emp_cov.copy()
    covariance_ *= 0.95
    covariance_.flat[::n_features + 1] = emp_cov.flat[::n_features + 1]
    from scipy import linalg
    Z = linalg.pinvh(covariance_)

    # Z = np.zeros_like(emp_cov)
    U = np.zeros_like(emp_cov)
    Z_old = np.zeros_like(Z)

    checks = []
    for iteration_ in range(max_iter):
        # x-update
        A = Z - U
        A += A.T
        A /= 2.
        K = prox_logdet(emp_cov - rho * A, lamda=1. / rho)

        # z-update with relaxation
        K_hat = over_relax * K - (1 - over_relax) * Z
        Z = soft_thresholding_od(K_hat + U, lamda=alpha / rho)

        # update residuals
        U += K_hat - Z

        # diagnostics, reporting, termination checks
        obj = objective(emp_cov, K, Z, alpha) if compute_objective else np.nan
        rnorm = np.linalg.norm(K - Z, 'fro')
        snorm = rho * np.linalg.norm(Z - Z_old, 'fro')
        check = convergence(
            obj=obj, rnorm=rnorm, snorm=snorm,
            e_pri=np.sqrt(K.size) * tol + rtol * max(
                np.linalg.norm(K, 'fro'), np.linalg.norm(Z, 'fro')),
            e_dual=np.sqrt(K.size) * tol + rtol * rho * np.linalg.norm(U))

        Z_old = Z.copy()
        if verbose:
            print(
                "obj: %.4f, rnorm: %.4f, snorm: %.4f,"
                "eps_pri: %.4f, eps_dual: %.4f" % check)

        checks.append(check)
        if check.rnorm <= check.e_pri and check.snorm <= check.e_dual:
            break

        rho_new = update_rho(
            rho, rnorm, snorm, iteration=iteration_,
            **(update_rho_options or {}))
        # scaled dual variables should be also rescaled
        U *= rho / rho_new
        rho = rho_new
    else:
        warnings.warn("Objective did not converge.")

    return_list = [Z, emp_cov]
    if return_history:
        return_list.append(checks)
    if return_n_iter:
        return_list.append(iteration_)
    return return_list


class GraphLasso(GraphLasso):
    """Sparse inverse covariance estimation with an l1-penalized estimator.

    Parameters
    ----------
    alpha : positive float, default 0.01
        The regularization parameter: the higher alpha, the more
        regularization, the sparser the inverse covariance.

    rho : positive float, default 1
        Augmented Lagrangian parameter.

    over_relax : positive float, deafult 1
        Over-relaxation parameter (typically between 1.0 and 1.8).

    tol : positive float, default 1e-4
        Absolute tolerance to declare convergence.

    rtol : positive float, default 1e-4
        Relative tolerance to declare convergence.

    max_iter : integer, default 100
        The maximum number of iterations.

    verbose : boolean, default False
        If verbose is True, the objective function, rnorm and snorm are
        printed at each iteration.

    assume_centered : boolean, default False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    update_rho_options : dict, default None
        Options for the update of rho. See `update_rho` function for details.

    compute_objective : boolean, default True
        Choose if compute the objective function during iterations
        (only useful if `verbose=True`).

    mode : {'admm'}, default 'admm'
        Minimisation algorithm. At the moment, only 'admm' is available,
        so this is ignored.

    Attributes
    ----------
    covariance_ : array-like, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : array-like, shape (n_features, n_features)
        Estimated pseudo inverse matrix.

    n_iter_ : int
        Number of iterations run.

    """

    def __init__(
            self, alpha=0.01, rho=1., over_relax=1., max_iter=100, mode='admm',
            tol=1e-4, rtol=1e-4, verbose=False, assume_centered=False,
            update_rho_options=None, compute_objective=True):
        super(GraphLasso, self).__init__(
            alpha=alpha, tol=tol, max_iter=max_iter, verbose=verbose,
            assume_centered=assume_centered, mode=mode)
        self.rho = rho
        self.rtol = rtol
        self.over_relax = over_relax
        self.update_rho_options = update_rho_options
        self.compute_objective = compute_objective

    def _fit(self, emp_cov):
        """Fit the GraphLasso model to X.

        Parameters
        ----------
        emp_cov : ndarray, shape (n_features, n_features)
            Empirical covariance of data.

        """
        self.precision_, self.covariance_, self.n_iter_ = graph_lasso(
            emp_cov, alpha=self.alpha, tol=self.tol, rtol=self.rtol,
            max_iter=self.max_iter, over_relax=self.over_relax, rho=self.rho,
            verbose=self.verbose, return_n_iter=True, return_history=False,
            mode=self.mode, update_rho_options=self.update_rho_options,
            compute_objective=self.compute_objective)
        return self

    def fit(self, X, y=None):
        """Fit the GraphLasso model to X.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the covariance estimate
        y : (ignored)

        """
        # Covariance does not make sense for a single feature
        X = check_array(
            X, ensure_min_features=2, ensure_min_samples=2, estimator=self)
        if self.assume_centered:
            self.location_ = np.zeros(X.shape[1])
        else:
            self.location_ = X.mean(0)

        emp_cov = empirical_covariance(X, assume_centered=self.assume_centered)
        return self._fit(emp_cov)



import numpy as np
from scipy import linalg
from sklearn.utils.extmath import fast_logdet


def h(D, K, delta=5):
    return - 0.5 * (np.trace(K.T.dot(D)) - (delta - 2) * fast_logdet(K))


def first_derivative_h(D, K, delta=5):
    return - 0.5 * (D - (delta - 2) * linalg.pinvh(K))


# def second_derivative_h(D, K, delta=5):
#     K_inv = linalg.pinvh(K)
#     return - 0.5 * (delta - 2) * K_inv.dot(K_inv)


def first_derivative_h_version2(D, K, delta=5):
    res = np.empty_like(D)
    for i in range(D.shape[0]):
        for j in range(D.shape[1]):
            ones = np.zeros_like(D)
            ones[i, j] = ones[j, i] = 1.
            res[i, j] = np.trace((D - (delta - 2) * linalg.pinvh(K)).dot(ones))
    return - res / 2.


def second_derivative_h_version2(D, K, delta=5):
    K_inv = linalg.pinvh(K)
    num_edges = np.triu(K != 0, 1).sum()
    dof = num_edges + K.shape[0]
    H = np.empty((dof, dof))
    # H2 = np.empty((dof, dof))
    Vi, Vj = np.nonzero(np.triu(K != 0))

    # to be the same as matlab
    idx = np.argsort(Vj)
    Vi, Vj = Vi[idx], Vj[idx]

    for ei in range(dof):
        i, j = Vi[ei], Vj[ei]

        # ones1 = np.zeros_like(K)
        # ones1[i, j] = ones1[j, i] = 1.
        # A2 = K_inv.dot(ones1)

        for ej in range(ei, dof):
            l, m = Vi[ej], Vj[ej]
            # ones2 = np.zeros_like(K)
            # ones2[l, m] = ones2[m, l] = 1.

            A = K_inv[[l, m]][:, [j, i]]
            B = K_inv[[i, j]][:, [m, l]]
            # B2 = K_inv.dot(ones2)

            # const = 1
            # if i == j:
            #     const *= 2
            # if l == m:
            #     const *= 2

            # assert np.allclose((A * B.T).sum(), const * (A2 * B2.T).sum())
            H[ei, ej] = H[ej, ei] = (A * B.T).sum()
            # H2[ei, ej] = H2[ej, ei] = (A2 * B2.T).sum()
    return - H * (delta - 2) / 2.


# if __name__ == '__main__':
#     K = make_sparse_spd_matrix(4)
#     D = make_spd_matrix(4)
#
#     assert np.allclose(first_derivative_h(D, K),
#                        first_derivative_h_version2(D, K))



"""Useful norms."""
import numpy as np
from scipy.optimize import minimize


def vector_p_norm(a, p=1, time_on_axis='first'):
    """Sum of norms for each vector."""
    b = np.array([b.flatten() for b in a]).T
    return np.linalg.norm(b, axis=1, ord=1).sum()
    # if time_on_axis == 'first':
    #     a = a.T
    # return np.sum(np.linalg.norm(a.reshape(np.prod(a.shape[:-1]), a.shape[-1]),
    #               axis=1, ord=p))


def l1_norm(precision):
    """L1 norm."""
    return np.abs(precision).sum()


def l1_od_norm(precision):
    """L1 norm off-diagonal."""
    return np.abs(precision).sum() - np.abs(np.diag(precision)).sum()


def node_penalty(X):
    """Node penalty. See Hallac for details."""
    cons = (
        {'type': 'eq',
         'fun': lambda x: np.array((x.reshape(X.shape) + x.reshape(X.shape).T - X).sum()), 'jac': lambda x: np.full(np.prod(X.shape), 2)},
        # {'type': 'ineq',  # avoid problem with scipy
        #  'fun': lambda x: 1,
        #  'jac': lambda x: np.zeros(np.prod(X.shape))}
    )
    try:
        res = minimize(
            lambda x: np.sum(np.linalg.norm(x.reshape(X.shape), axis=0)),
            np.random.randn(np.prod(X.shape)), constraints=cons).fun
    except ValueError:
        res = np.nan

    return res



"""Sampling module for Wishart processes."""
from functools import partial

import numpy as np
from scipy import linalg, stats
from sklearn.utils.extmath import fast_logdet
from sklearn.covariance import empirical_covariance, log_likelihood

from regain.bayesian.stats import (
    log_likelihood_normal, lognormal_logpdf, lognormal_pdf, lognstat)
from regain.bayesian.wishart_process_ import GWP_construct


def elliptical_slice(
        xx, prior, cur_log_like, likelihood=None, angle_range=0, max_iter=20):
    """Markov chain update for a distribution with a Gaussian "prior" factored out.

    A Markov chain update is applied to the D-element array xx leaving a
    "posterior" distribution
        P(xx) \propto N(xx0,Sigma) \ell(xx)
    invariant. Where N(0,Sigma) is a zero-mean Gaussian distribution with
    covariance Sigma. Often \ell is a likelihood function.

    Parameters
    ----------
    xx : array-like, shape (D,)
        Initial vector.
    prior :  array-like, shape (D,)
        Single sample from N(0, Sigma)
    cur_log_like : float
        Current log-likelihood.
    angle_range : float, default 0
        Explore whole ellipse with break point at first rejection.
        Set in (0,2*pi] to explore a bracket of the specified width
        centred uniformly at random.

    Returns:
    --------
    xx : np.ndarray, shape (D,)
        Perturbed vector.
    cur_log_like : float
        Log-likelihood of xx.

    Originally written in matlab by Iain Murray
    http://homepages.inf.ed.ac.uk/imurray2/pub/10ess/elliptical_slice.m
    Iain Murray, September 2009
    Tweak to interface and documentation, September 2010

    Reference
    ---------
    Elliptical slice sampling
    Iain Murray, Ryan Prescott Adams and David J.C. MacKay.
    The Proceedings of the 13th International Conference on Artificial
    Intelligence and Statistics (AISTATS), JMLR W&CP 9:541-548, 2010.
    """
    if likelihood is None:
        raise ValueError(
            "`likelihood` parameter is None, should be a "
            "function to evaluate likelihood")
    initial_theta = xx.xx
    v, p, N = initial_theta.shape
    D = v * p * N

    # S = xx.S
    L = xx.L

    cur_log_like_start = cur_log_like
    if cur_log_like is None:
        # cur_log_like = log_lik_frob(S, xx.V, variance)
        # cur_log_like = time_multivariate_normal_logpdf(initial_theta, xx.V)
        cur_log_like = likelihood(xx.V)

    # Set up the ellipse and the slice threshold
    if prior.size == D:
        #  User provided a prior sample:
        nu = prior
    else:
        #  User specified Cholesky of prior covariance:
        if prior.shape != (D, D):
            raise ValueError(
                'Prior must be given by a D-element sample '
                'or DxD chol(Sigma)')
        nu = np.reshape(
            prior.T.dot(np.random.normal(size=D)), initial_theta.shape)

    hh = 0.001 * np.log(np.random.uniform()) + cur_log_like

    #  Set up a bracket of angles and pick a first proposal.
    #  "phi = (theta'-theta)" is a change in angle.
    if angle_range <= 0:
        #  Bracket whole ellipse with both edges at first proposed point
        phi = np.random.uniform() * 2 * np.pi
        phi_min = phi - 2 * np.pi
        phi_max = phi
    else:
        #  Randomly center bracket on current point
        phi_min = -angle_range * np.random.uniform()
        phi_max = phi_min + angle_range
        phi = np.random.uniform() * (phi_max - phi_min) + phi_min

    # Slice sampling loop
    error = False
    # LLt = L.dot(L.T)
    for iteration_ in range(max_iter):
        # Compute xx for proposed angle difference and check if on the slice
        xx_proposal = np.real(initial_theta * np.cos(phi) + nu * np.sin(phi))

        uut = np.array([u.dot(u.T) for u in xx_proposal.T])
        V = GWP_construct(xx_proposal, L, uut=uut)
        # cur_log_like = log_lik_frob(S, V, variance)
        # cur_log_like = stats.wishart.logpdf(V, nu, LLt)
        # cur_log_like = time_multivariate_normal_logpdf(xx_proposal, V)
        cur_log_like = likelihood(V)

        if cur_log_like > hh:
            # New point is on slice, ** EXIT LOOP **
            break

        # Shrink slice to rejected point
        if phi > 0:
            phi_max = phi
        elif phi < 0:
            phi_min = phi
        else:
            # error = True
            # break
            raise RuntimeError(
                'BUG DETECTED: Shrunk to current position '
                'and still not acceptable.')

        # Propose new angle difference
        phi = np.random.uniform() * (phi_max - phi_min) + phi_min
    else:
        error = True

    if error:
        # revert to initial point
        xx['xx'] = initial_theta
    else:
        # update with new point
        xx['uut'] = uut
        xx['xx'] = xx_proposal
        xx['V'] = V

    # return xx, cur_log_like_start if error else cur_log_like
    return xx, cur_log_like_start if error else cur_log_like


def sample_hyper_kernel(
        initial_theta, var_proposal, t, u, kern, mean_prior, var_prior):
    """Metropolis-Hastings for sampling the posterior of the kernel
    hyperparameter.

    According to the paper, we use a lognormal distribution as the proposal.

    Parameters
    ----------
    initial_theta : type
        Initial kernel hyperparameter.
    var_proposal : type
        Variance for the proposal.
    t : type
        Description of parameter `t`.
    u : type
        Description of parameter `u`.
    kern : type
        Description of parameter `kern`.
    mean_prior : type
        Description of parameter `mean_prior`.
    var_prior : type
        Description of parameter `var_prior`.

    Returns
    -------
    type
        Description of returned object.

    """
    # Propose a sample
    mu, sigma = lognstat(initial_theta, var_proposal)
    proposal = np.random.lognormal(mu, sigma)

    # Criterion to choose whether to accept the proposed sample or not
    logpzast = logpunorm(proposal, t, u, kern, mean_prior, var_prior)
    qzastztau = lognormal_pdf(proposal, mu=mu, sigma=sigma)

    mu, sigma = lognstat(proposal, var_proposal)
    logpztau = logpunorm(initial_theta, t, u, kern, mean_prior, var_prior)
    qztauzast = lognormal_pdf(initial_theta, mu=mu, sigma=sigma)

    acceptance_proba = min(
        1, np.exp(logpzast - logpztau) * (qztauzast / qzastztau))

    # Now we decide whether to accept zast or use the previous value
    accept = np.random.uniform() < acceptance_proba
    sample = proposal if accept else initial_theta
    return sample, accept


def logpunorm(inverse_width, t, u, kern, mean_prior, var_prior):
    """Posterior probability of inverse_width.

    Parameters
    ----------
    inverse_width : float
        Kernel parameter.
    t : ndarray
        Points where to compute the kernel.
    u : ndarray, shape (v, p, n)
        Sample tensor.
    kern : function
        Function for computing the kernel.
    mean_prior : float
        Prior for the mean.
    var_prior : float
        Prior for the variance.

    Returns
    -------
    logprob
        Posterior probability.
    """
    K = kern(t[:, None], inverse_width=inverse_width)
    k_inverse = linalg.pinvh(K)

    v, p, n = u.shape
    # F = np.tensordot(u, u, axes=([1, 0], [1, 0]))
    # logpugl = v * p * fast_logdet(k_inverse) - np.sum(F * k_inverse)
    # logpugl -= v * p * n * np.log(2 * np.pi)
    # logpugl /= 2.

    # creates a (vp * n) matrix, then compute centered empirical covariance
    ustack = np.vstack(u)
    logpugl = v * p * log_likelihood(empirical_covariance(ustack), k_inverse)

    mu_prior, sigma_prior = lognstat(mean_prior, var_prior)
    logp_prior = lognormal_logpdf(
        inverse_width, mu=mu_prior, sigma=sigma_prior)

    logprob = logpugl + logp_prior
    return logprob


def sample_ell(
        Ltau, var_proposal, umat, mu_prior, var_prior, uut=None,
        likelihood=None):
    """Metropolis-Hastings for sampling the posterior of the elements in L.

    Use a spherical normal distribution as the proposal.
    """
    # Run the MH individually per component of L
    free_elements = Ltau.size
    L_proposal = np.zeros(free_elements)

    if not isinstance(var_proposal, np.ndarray):
        var_proposal = var_proposal * np.ones(free_elements)
    if not isinstance(mu_prior, np.ndarray):
        mu_prior = mu_prior * np.ones(free_elements)
    if not isinstance(var_prior, np.ndarray):
        var_prior = var_prior * np.ones(free_elements)

    for i in range(free_elements):
        L_proposal[i] = _sample_ell_comp(
            Ltau, i, var_proposal[i], umat, mu_prior=mu_prior[i],
            var_prior=var_prior[i], uut=uut, likelihood=likelihood)
        Ltau[i] = L_proposal[i]

    return L_proposal


def _sample_ell_comp(
        Ltaug, i, sigma2Lprop, umat, mu_prior, var_prior, uut=None,
        likelihood=None):
    """Sample a single element for L."""
    if likelihood is None:
        raise ValueError(
            "`likelihood` parameter is None, should be a "
            "function to evaluate likelihood")
    # Propose a sample
    Ltau = Ltaug[i]
    Last = np.random.normal(Ltau, np.sqrt(sigma2Lprop))
    Lastg = Ltaug.copy()
    Lastg[i] = Last

    # Criterion to choose whether to accept the proposed sample or not
    # normpdf = lambda x, m, s: ...
    # np.exp(-0.5 * ((x - m)/s)**2) / (np.sqrt(2*np.pi) * s)

    logp_post = partial(
        logp_ell_posterior, i=i, u=umat, mu_prior=mu_prior,
        var_prior=var_prior, uut=uut, likelihood=likelihood)

    logp_diff = logp_post(Lastg) - logp_post(Ltaug)
    q_ast_tau = stats.norm.pdf(Last, Ltau, np.sqrt(sigma2Lprop))
    q_tau_ast = stats.norm.pdf(Ltau, Last, np.sqrt(sigma2Lprop))

    # Now we decide whether to accept zast or use the previous value
    accept = min(1, np.exp(logp_diff) * (q_tau_ast / q_ast_tau))
    return Last if np.random.uniform() < accept else Ltau


def logp_ell_posterior(
        ell_lower, i, u, mu_prior, var_prior, uut=None, likelihood=None):
    """Log-probability of the posterior of L.

    Parameters
    ----------
    ell_lower : ndarray
        Lower Cholesky.
    i : type
        Index.
    u : type
        Description of parameter `u`.
    mu_prior : type
        Description of parameter `mu_prior`.
    var_prior : type
        Description of parameter `var_prior`.
    uut : type
        Description of parameter `uut` (the default is None).
    likelihood : function
        Likelihood function (takes one parameter) (the default is None).

    Returns
    -------
    type
        Description of returned object.

    """
    v, p, n = u.shape
    ell = np.zeros((p, p))
    ell[np.tril_indices_from(ell)] = ell_lower
    D = GWP_construct(u, ell, uut=uut)
    # logpS = log_lik_frob(S, D, var_err)
    posterior = likelihood(D)
    prior = log_likelihood_normal(ell_lower[i], mu_prior, var_prior)
    return posterior + prior



"""Graphical latent variable models selection over time via ADMM."""
from __future__ import division

import warnings
from functools import partial

import numpy as np
import scipy.sparse as sp
from six.moves import map, range, zip
from sklearn.utils import check_array
from sklearn.utils.extmath import squared_norm

from regain.covariance.latent_time_graph_lasso_ import LatentTimeGraphLasso
from regain.norm import l1_od_norm
from regain.prox import prox_trace_indicator
from regain.prox import soft_thresholding_sign as soft_thresholding
from regain.update_rules import update_rho
from regain.utils import convergence
from regain.validation import check_array_dimensions, check_norm_prox


def objective(S, R, Z_0, Z_1, Z_2, W_0, W_1, W_2,
              alpha, tau, beta, eta, psi, phi):
    """Objective function for latent variable time-varying graphical lasso."""
    obj = squared_norm(S - R)
    obj += alpha * sum(map(l1_od_norm, Z_0))
    obj += tau * sum(map(partial(np.linalg.norm, ord='nuc'), W_0))
    obj += beta * sum(map(psi, Z_2 - Z_1))
    obj += eta * sum(map(phi, W_2 - W_1))
    return obj


def latent_time_matrix_decomposition(
        emp_cov, alpha=0.01, tau=1., rho=1., beta=1., eta=1., max_iter=100,
        verbose=False, psi='laplacian', phi='laplacian', mode='admm',
        tol=1e-4, rtol=1e-4, assume_centered=False,
        return_history=False, return_n_iter=True,
        update_rho_options=None, compute_objective=True):
    r"""Latent variable time-varying matrix decomposition solver.

    Solves the following problem via ADMM:
        min sum_{i=1}^T || S_i-(K_i-L_i)||^2 + alpha ||K_i||_{od,1}
            + tau ||L_i||_*
            + beta sum_{i=2}^T Psi(K_i - K_{i-1})
            + eta sum_{i=2}^T Phi(L_i - L_{i-1})

    where S is the matrix to decompose.

    Parameters
    ----------
    emp_cov : ndarray, shape (n_features, n_features)
        Matrix to decompose.
    alpha, tau, beta, eta : float, optional
        Regularisation parameters.
    rho : float, optional
        Augmented Lagrangian parameter.
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.

    Returns
    -------
    K, L : numpy.array, 3-dimensional (T x d x d)
        Solution to the problem for each time t=1...T .
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    psi, prox_psi, psi_node_penalty = check_norm_prox(psi)
    phi, prox_phi, phi_node_penalty = check_norm_prox(phi)

    Z_0 = np.zeros_like(emp_cov)
    Z_1 = np.zeros_like(Z_0)[:-1]
    Z_2 = np.zeros_like(Z_0)[1:]
    W_0 = np.zeros_like(Z_0)
    W_1 = np.zeros_like(Z_1)
    W_2 = np.zeros_like(Z_2)

    X_0 = np.zeros_like(Z_0)
    X_1 = np.zeros_like(Z_1)
    X_2 = np.zeros_like(Z_2)
    U_1 = np.zeros_like(W_1)
    U_2 = np.zeros_like(W_2)

    R_old = np.zeros_like(Z_0)
    Z_1_old = np.zeros_like(Z_1)
    Z_2_old = np.zeros_like(Z_2)
    W_1_old = np.zeros_like(W_1)
    W_2_old = np.zeros_like(W_2)

    # divisor for consensus variables, accounting for two less matrices
    divisor = np.full(emp_cov.shape[0], 3, dtype=float)
    divisor[0] -= 1
    divisor[-1] -= 1

    checks = []
    for iteration_ in range(max_iter):
        # update R
        A = Z_0 - W_0 - X_0
        R = (rho * A + 2 * emp_cov) / (2 + rho)

        # update Z_0
        A = R + W_0 + X_0
        A[:-1] += Z_1 - X_1
        A[1:] += Z_2 - X_2
        A /= divisor[:, None, None]
        # soft_thresholding_ = partial(soft_thresholding, lamda=alpha / rho)
        # Z_0 = np.array(map(soft_thresholding_, A))
        Z_0 = soft_thresholding(
            A, lamda=alpha / (rho * divisor[:, None, None]))

        # update Z_1, Z_2
        A_1 = Z_0[:-1] + X_1
        A_2 = Z_0[1:] + X_2
        if not psi_node_penalty:
            prox_e = prox_psi(A_2 - A_1, lamda=2. * beta / rho)
            Z_1 = .5 * (A_1 + A_2 - prox_e)
            Z_2 = .5 * (A_1 + A_2 + prox_e)
        else:
            Z_1, Z_2 = prox_psi(np.concatenate((A_1, A_2), axis=1),
                                lamda=.5 * beta / rho,
                                rho=rho, tol=tol, rtol=rtol, max_iter=max_iter)

        # update W_0
        A = Z_0 - R - X_0
        A[:-1] += W_1 - U_1
        A[1:] += W_2 - U_2
        A /= divisor[:, None, None]
        A += A.transpose(0, 2, 1)
        A /= 2.

        W_0 = np.array([prox_trace_indicator(a, lamda=tau / (rho * div))
                        for a, div in zip(A, divisor)])

        # update W_1, W_2
        A_1 = W_0[:-1] + U_1
        A_2 = W_0[1:] + U_2
        if not phi_node_penalty:
            prox_e = prox_phi(A_2 - A_1, lamda=2. * eta / rho)
            W_1 = .5 * (A_1 + A_2 - prox_e)
            W_2 = .5 * (A_1 + A_2 + prox_e)
        else:
            W_1, W_2 = prox_phi(np.concatenate((A_1, A_2), axis=1),
                                lamda=.5 * eta / rho,
                                rho=rho, tol=tol, rtol=rtol, max_iter=max_iter)

        # update residuals
        X_0 += R - Z_0 + W_0
        X_1 += Z_0[:-1] - Z_1
        X_2 += Z_0[1:] - Z_2
        U_1 += W_0[:-1] - W_1
        U_2 += W_0[1:] - W_2

        # diagnostics, reporting, termination checks
        rnorm = np.sqrt(
            squared_norm(R - Z_0 + W_0) +
            squared_norm(Z_0[:-1] - Z_1) + squared_norm(Z_0[1:] - Z_2) +
            squared_norm(W_0[:-1] - W_1) + squared_norm(W_0[1:] - W_2))

        snorm = rho * np.sqrt(
            squared_norm(R - R_old) +
            squared_norm(Z_1 - Z_1_old) + squared_norm(Z_2 - Z_2_old) +
            squared_norm(W_1 - W_1_old) + squared_norm(W_2 - W_2_old))

        obj = objective(emp_cov, R, Z_0, Z_1, Z_2, W_0, W_1, W_2,
                        alpha, tau, beta, eta, psi, phi) \
            if compute_objective else np.nan

        check = convergence(
            obj=obj, rnorm=rnorm, snorm=snorm,
            e_pri=np.sqrt(R.size + 4 * Z_1.size) * tol + rtol * max(
                np.sqrt(squared_norm(R) +
                        squared_norm(Z_1) + squared_norm(Z_2) +
                        squared_norm(W_1) + squared_norm(W_2)),
                np.sqrt(squared_norm(Z_0 - W_0) +
                        squared_norm(Z_0[:-1]) + squared_norm(Z_0[1:]) +
                        squared_norm(W_0[:-1]) + squared_norm(W_0[1:]))),
            e_dual=np.sqrt(R.size + 4 * Z_1.size) * tol + rtol * rho * (
                np.sqrt(squared_norm(X_0) +
                        squared_norm(X_1) + squared_norm(X_2) +
                        squared_norm(U_1) + squared_norm(U_2))))

        R_old = R.copy()
        Z_1_old = Z_1.copy()
        Z_2_old = Z_2.copy()
        W_1_old = W_1.copy()
        W_2_old = W_2.copy()

        if verbose:
            print("obj: %.4f, rnorm: %.4f, snorm: %.4f,"
                  "eps_pri: %.4f, eps_dual: %.4f" % check)

        checks.append(check)
        if check.rnorm <= check.e_pri and check.snorm <= check.e_dual:
            break

        rho_new = update_rho(rho, rnorm, snorm, iteration=iteration_,
                             **(update_rho_options or {}))
        # scaled dual variables should be also rescaled
        X_0 *= rho / rho_new
        X_1 *= rho / rho_new
        X_2 *= rho / rho_new
        U_1 *= rho / rho_new
        U_2 *= rho / rho_new
        rho = rho_new
    else:
        warnings.warn("Objective did not converge.")

    return_list = [Z_0, W_0]
    if return_history:
        return_list.append(checks)
    if return_n_iter:
        return_list.append(iteration_)
    return return_list


class LatentTimeMatrixDecomposition(LatentTimeGraphLasso):
    """Sparse inverse covariance estimation with an l1-penalized estimator.

    Parameters
    ----------
    alpha : positive float, default 0.01
        Regularization parameter for precision matrix. The higher alpha,
        the more regularization, the sparser the inverse covariance.

    tau : positive float, default 1
        Regularization parameter for latent variables matrix. The higher tau,
        the more regularization, the lower rank of the latent matrix.

    beta : positive float, default 1
        Regularization parameter to constrain precision matrices in time.
        The higher beta, the more regularization,
        and consecutive precision matrices in time are more similar.

    psi : {'laplacian', 'l1', 'l2', 'linf', 'node'}, default 'laplacian'
        Type of norm to enforce for consecutive precision matrices in time.

    eta : positive float, default 1
        Regularization parameter to constrain latent matrices in time.
        The higher eta, the more regularization,
        and consecutive latent matrices in time are more similar.

    phi : {'laplacian', 'l1', 'l2', 'linf', 'node'}, default 'laplacian'
        Type of norm to enforce for consecutive latent matrices in time.

    rho : positive float, default 1
        Augmented Lagrangian parameter.

    over_relax : positive float, deafult 1
        Over-relaxation parameter (typically between 1.0 and 1.8).

    tol : positive float, default 1e-4
        Absolute tolerance to declare convergence.

    rtol : positive float, default 1e-4
        Relative tolerance to declare convergence.

    max_iter : integer, default 100
        The maximum number of iterations.

    verbose : boolean, default False
        If verbose is True, the objective function, rnorm and snorm are
        printed at each iteration.

    assume_centered : boolean, default False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    time_on_axis : {'first', 'last'}, default 'first'
        If data have time as the last dimension, set this to 'last'.
        Useful to use scikit-learn functions as train_test_split.

    update_rho_options : dict, default None
        Options for the update of rho. See `update_rho` function for details.

    compute_objective : boolean, default True
        Choose if compute the objective function during iterations
        (only useful if `verbose=True`).

    mode : {'admm'}, default 'admm'
        Minimisation algorithm. At the moment, only 'admm' is available,
        so this is ignored.

    Attributes
    ----------
    precision_ : array-like, shape (n_times, n_features, n_features)
        Estimated pseudo inverse matrix.

    latent_ : array-like, shape (n_times, n_features, n_features)
        Estimated latent variable matrix.

    n_iter_ : int
        Number of iterations run.

    """

    def __init__(self, alpha=0.01, tau=1., beta=1., eta=1., mode='admm', rho=1.,
                 time_on_axis='first', tol=1e-4, rtol=1e-4,
                 psi='laplacian', phi='laplacian', max_iter=100,
                 verbose=False, assume_centered=False, update_rho_options=None,
                 compute_objective=True):
        super(LatentTimeMatrixDecomposition, self).__init__(
            alpha=alpha, beta=beta, tau=tau, eta=eta,
            mode=mode, rho=rho, tol=tol, rtol=rtol,
            psi=psi, phi=phi, max_iter=max_iter, verbose=verbose,
            time_on_axis=time_on_axis,
            assume_centered=assume_centered,
            update_rho_options=update_rho_options,
            compute_objective=compute_objective)

    def _fit(self, X):
        """Fit the LatentTimeMatrixDecomposition model to X.

        Parameters
        ----------
        X : ndarray, shape (n_time, n_samples, n_features), or
                (n_samples, n_features, n_time)
            Matrix to decompose.

        """
        self.precision_, self.latent_, self.n_iter_ = \
            latent_time_matrix_decomposition(
                X, alpha=self.alpha, tau=self.tau, rho=self.rho,
                beta=self.beta, eta=self.eta, mode=self.mode,
                tol=self.tol, rtol=self.rtol, psi=self.psi, phi=self.phi,
                max_iter=self.max_iter, verbose=self.verbose,
                return_n_iter=True, return_history=False,
                update_rho_options=self.update_rho_options,
                compute_objective=self.compute_objective)
        self.reconstruction_err_ = squared_norm(X - self.precision_ + self.latent_)
        return self

    def fit(self, X, y=None):
        """Fit the LatentTimeMatrixDecomposition model to X.

        Parameters
        ----------
        X : ndarray, shape (n_time, n_samples, n_features), or
                (n_samples, n_features, n_time)
            Matrix to decompose.
            If shape is (n_samples, n_features, n_time), then set
            `time_on_axis = 'last'`.
        y : (ignored)

        """
        if sp.issparse(X):
            raise TypeError("sparse matrices not supported.")

        X = check_array_dimensions(
            X, n_dimensions=3, time_on_axis=self.time_on_axis)

        # Covariance does not make sense for a single feature
        X = np.array([check_array(x, ensure_min_features=2,
                      ensure_min_samples=2, estimator=self) for x in X])

        return self._fit(X)



"""Time graph lasso via forward_backward (for now only in case of l1 norm)."""
from __future__ import division, print_function

import warnings
from functools import partial

import numpy as np
from scipy import linalg
from six.moves import map, range, zip
from sklearn.covariance import empirical_covariance

from regain.covariance.graph_lasso_ import logl
from regain.covariance.time_graph_lasso_ import TimeGraphLasso
from regain.norm import l1_od_norm, vector_p_norm
from regain.prox import prox_FL
from regain.utils import convergence, positive_definite


def loss(S, K, n_samples=None, vareps=0):
    """Loss function for time-varying graphical lasso."""
    if n_samples is None:
        n_samples = np.ones(S.shape[0])
    loss_ = sum(-ni * logl(emp_cov, precision)
                for emp_cov, precision, ni in zip(S, K, n_samples))
    # loss_ += vareps / 2. * squared_norm(K)
    loss_ += vareps / 2. * _scalar_product(K, K)
    return loss_


def grad_loss(x, emp_cov, n_samples, x_inv=None, vareps=0):
    """Gradient of the loss function for the time-varying graphical lasso."""
    if x_inv is None:
        x_inv = np.array([linalg.pinvh(_) for _ in x])
    grad = emp_cov - x_inv
    grad *= n_samples[:, None, None]

    # add coercitive term
    grad += vareps * x

    return grad


def penalty(precision, alpha, beta, psi):
    """Penalty for time-varying graphical lasso."""
    if isinstance(alpha, np.ndarray):
        obj = sum(a[0][0] * m for a, m in zip(alpha, map(l1_od_norm, precision)))
    else:
        obj = alpha * sum(map(l1_od_norm, precision))
    obj += beta * psi(precision[1:] - precision[:-1])
    return obj


def objective(n_samples, emp_cov, precision, alpha, beta, psi, vareps=0):
    """Objective function for time-varying graphical lasso."""
    obj = loss(emp_cov, precision, n_samples=n_samples, vareps=vareps)
    obj += penalty(precision, alpha, beta, psi)
    return obj


def _J(x, beta, alpha, gamma, lamda, S, n_samples, p=1, x_inv=None, grad=None):
    """Grad + prox + line search for the new point."""
    # if grad is None:
    #     grad = grad_loss(x, S, n_samples, x_inv=x_inv)
    prox = prox_FL(
        x - gamma * grad, beta * gamma, alpha * gamma, p=p, symmetric=True)
    return x + lamda * (prox - x)


def _scalar_product(x, y):
    return (x * y).sum()
    # return x.ravel().dot(y.ravel())


def choose_gamma(gamma, x, emp_cov, n_samples, beta, alpha, lamda, grad,
                 delta=1e-4, eps=0.5, max_iter=1000, p=1, x_inv=None,
                 vareps=1e-5, choose='gamma'):
    """Choose gamma for backtracking.

    References
    ----------
    Salzo S. (2017). https://doi.org/10.1137/16M1073741

    """
    # if grad is None:
    #     grad = grad_loss(x, emp_cov, n_samples, x_inv=x_inv)

    partial_f = partial(loss, n_samples=n_samples, S=emp_cov, vareps=vareps)
    fx = partial_f(K=x)
    for i in range(max_iter):
        prox = prox_FL(
            x - gamma * grad, beta * gamma, alpha * gamma, p=p, symmetric=True)
        if positive_definite(prox) and choose != "gamma":
            break

        if choose == "gamma":
            y_minus_x = prox - x
            loss_diff = partial_f(K=x + lamda * y_minus_x) - fx

            tolerance = _scalar_product(y_minus_x, grad)
            tolerance += delta / gamma * _scalar_product(y_minus_x, y_minus_x)
            if loss_diff <= lamda * tolerance:
                break
        gamma *= eps

    return gamma, prox


def choose_lamda(lamda, x, emp_cov, n_samples, beta, alpha, gamma, delta=1e-4,
                 eps=0.5, max_iter=1000, criterion='b', p=1, x_inv=None,
                 grad=None, prox=None, min_eigen_x=None,
                 vareps=1e-5):
    """Choose lambda for backtracking.

    References
    ----------
    Salzo S. (2017). https://doi.org/10.1137/16M1073741

    """
    # if x_inv is None:
    #     x_inv = np.array([linalg.pinvh(_) for _ in x])
    # if grad is None:
    #     grad = grad_loss(x, emp_cov, n_samples, x_inv=x_inv)
    # if prox is None:
    #     prox = prox_FL(x - gamma * grad, beta * gamma, alpha * gamma, p=p, symmetric=True)

    partial_f = partial(loss, n_samples=n_samples, S=emp_cov, vareps=vareps)
    fx = partial_f(K=x)

    # min_eigen_y = np.min([np.linalg.eigh(z)[0] for z in prox])

    y_minus_x = prox - x
    if criterion == 'b':
        tolerance = _scalar_product(y_minus_x, grad)
        tolerance += delta / gamma * _scalar_product(y_minus_x, y_minus_x)
    elif criterion == 'c':
        psi = partial(vector_p_norm, p=p)
        gx = penalty(x, alpha, beta, psi)
        gy = penalty(prox, alpha, beta, psi)
        objective_x = objective(
            n_samples, emp_cov, x, alpha, beta, psi, vareps=vareps)
        tolerance = (1 - delta) * (gy - gx + _scalar_product(y_minus_x, grad))

    for i in range(max_iter):
        # line-search
        x1 = x + lamda * y_minus_x

        if criterion == 'a':
            iter_diff = x1 - x
            gradx1 = grad_loss(x1, emp_cov, n_samples)
            grad_diff = gradx1 - grad
            norm_grad_diff = np.sqrt(_scalar_product(grad_diff, grad_diff))
            norm_iter_diff = np.sqrt(_scalar_product(iter_diff, iter_diff))
            tolerance = delta * norm_iter_diff / (gamma * lamda)
            if norm_grad_diff <= tolerance:
                break
        elif criterion == 'b':
            loss_diff = partial_f(K=x1) - fx
            if loss_diff <= lamda * tolerance and positive_definite(x1):
                break
        elif criterion == 'c':
            obj_diff = objective(
                n_samples, emp_cov, x1, alpha, beta, psi, vareps=vareps) \
                    - objective_x
            # if positive_definite(x1) and obj_diff <= lamda * tolerance:
            cond = True # lamda > 0 if min_eigen_y >= 0 else lamda < min_eigen_x / (min_eigen_x - min_eigen_y)
            if cond and obj_diff <= lamda * tolerance:
                break
        else:
            raise ValueError(criterion)
        lamda *= eps
    return lamda, i + 1


def fista_step(Y, Y_diff, t):
    t_next = (1. + np.sqrt(1.0 + 4.0 * t*t)) / 2.
    return Y + ((t - 1.0)/t_next) * Y_diff, t_next


def upper_diag_3d(x):
    """Return the flattened upper diagonal of a 3d matrix."""
    # n_times, n_dim, _ = x.shape
    # upper_idx = np.triu_indices(n_dim, 1)
    # return np.array([xx[upper_idx] for xx in x])
    return np.triu(x, 1)


def time_graph_lasso(
        emp_cov, n_samples, alpha=0.01, beta=1., max_iter=100, verbose=False,
        tol=1e-4, delta=1e-4, gamma=1., lamda=1., eps=0.5, debug=False,
        return_history=False, return_n_iter=True, choose='gamma',
        lamda_criterion='b', time_norm=1, compute_objective=True,
        return_n_linesearch=False, vareps=1e-5, stop_at=None, stop_when=1e-4):
    """Time-varying graphical lasso solver.

    Solves the following problem via ADMM:
        minimize  trace(S*X) - log det X + lambda*||X||_1

    where S is the empirical covariance of the data
    matrix D (training observations by features).

    Parameters
    ----------
    data_list : list of 2-dimensional matrices.
        Input matrices.
    lamda : float, optional
        Regularisation parameter.
    rho : float, optional
        Augmented Lagrangian parameter.
    alpha : float, optional
        Over-relaxation parameter (typically between 1.0 and 1.8).
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.

    Returns
    -------
    X : numpy.array, 2-dimensional
        Solution to the problem.
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    available_choose = ('gamma', 'lamda', 'fixed', 'both')
    if choose not in available_choose:
        raise ValueError("`choose` parameter must be one of %s." %
                         available_choose)

    n_times, _, n_features = emp_cov.shape
    covariance_ = emp_cov.copy()
    covariance_ *= 0.95

    K = np.empty_like(emp_cov)
    for i, (c, e) in enumerate(zip(covariance_, emp_cov)):
        c.flat[::n_features + 1] = e.flat[::n_features + 1]
        K[i] = linalg.pinvh(c)

    # K = np.array([np.eye(s.shape[0]) for s in emp_cov])
    # Y = K.copy()
    # assert positive_definite(K)

    obj_partial = partial(
        objective, n_samples=n_samples, emp_cov=emp_cov,
        alpha=alpha, beta=beta, psi=partial(vector_p_norm, p=time_norm),
        vareps=vareps)

    max_residual = -np.inf
    n_linesearch = 0
    checks = [convergence(obj=obj_partial(precision=K))]
    for iteration_ in range(max_iter):
        # if not positive_definite(K):
        #     print("precision is not positive definite.")
        #     # break

        k_previous = K.copy()
        x_inv = np.array([linalg.pinvh(x) for x in K])
        # x_inv = []
        # eigens = []
        # for x in K:
        #     es, Q = np.linalg.eigh(x)
        #     Inv = np.linalg.multi_dot((Q, np.diag(1. / es), Q.T))
        #     x_inv.append(Inv)
        #     eigens.append(es)
        # x_inv = np.array(x_inv)
        # eigens = np.array(eigens)

        grad = grad_loss(K, emp_cov, n_samples, x_inv=x_inv, vareps=vareps)
        if choose in ['gamma', 'both']:
            gamma, y = choose_gamma(
                gamma / eps if iteration_ > 0 else gamma, K, emp_cov,
                n_samples=n_samples,
                beta=beta, alpha=alpha, lamda=lamda, grad=grad,
                delta=delta, eps=eps, max_iter=200, p=time_norm, x_inv=x_inv,
                vareps=vareps, choose=choose)
        # print(gamma)

        x_hat = K - gamma * grad
        if choose not in ['gamma', 'both']:
            y = prox_FL(x_hat, beta * gamma, alpha * gamma, p=time_norm, symmetric=True)

        if choose in ['lamda', 'both']:
            lamda, n_ls = choose_lamda(
                min(lamda / eps if iteration_ > 0 else lamda, 1),
                K, emp_cov, n_samples=n_samples,
                beta=beta, alpha=alpha, gamma=gamma, delta=delta, eps=eps,
                criterion=lamda_criterion, max_iter=200, p=time_norm,
                x_inv=x_inv, grad=grad, prox=y,
                # min_eigen_x=np.min(eigens),
                vareps=vareps)
            n_linesearch += n_ls
        # print ("lambda: ", lamda, n_ls)

        K = K + min(max(lamda, 0), 1) * (y - K)
        # K, t = fista_step(Y, Y - Y_old, t)

        check = convergence(
            obj=obj_partial(precision=K),
            rnorm=np.linalg.norm(upper_diag_3d(K) - upper_diag_3d(k_previous)),
            snorm=np.linalg.norm(
                obj_partial(precision=K) - obj_partial(precision=k_previous)),
            e_pri=np.sqrt(upper_diag_3d(K).size) * tol + tol * max(
                np.linalg.norm(upper_diag_3d(K)),
                np.linalg.norm(upper_diag_3d(k_previous))),
            e_dual=tol,
            # precision=K.copy()
            )

        if verbose and iteration_ % (50 if verbose < 2 else 1) == 0:
            print("obj: %.4f, rnorm: %.7f, snorm: %.4f,"
                  "eps_pri: %.4f, eps_dual: %.4f" % check[:5])

        if return_history:
            checks.append(check)

        if np.isnan(check.rnorm) or np.isnan(check.snorm):
            warnings.warn("precision is not positive definite.")

        if stop_at is not None:
            if abs(check.obj - stop_at) / abs(stop_at) < stop_when:
                break
        else:
            # use this convergence criterion
            subgrad = (x_hat - K) / gamma
            if 0:
                grad = grad_loss(K, emp_cov, n_samples, vareps=vareps)
                res_norm = np.linalg.norm(grad + subgrad)

                if iteration_ == 0:
                    normalizer = res_norm + 1e-6
                max_residual = max(np.linalg.norm(grad),
                                   np.linalg.norm(subgrad)) + 1e-6
            else:
                res_norm = np.linalg.norm(K - k_previous) / gamma
                max_residual = max(max_residual, res_norm)
                normalizer = max(np.linalg.norm(grad),
                                 np.linalg.norm(subgrad)) + 1e-6

            r_rel = res_norm / max_residual
            r_norm = res_norm / normalizer

            if not debug and (r_rel <= tol or r_norm <= tol) and iteration_ > 0: # or (
                    # check.rnorm <= check.e_pri and iteration_ > 0):
                break
            # pass
        # if check.rnorm <= check.e_pri and iteration_ > 0:
        #     # and check.snorm <= check.e_dual:
        #     break
    else:
        warnings.warn("Objective did not converge.")

    # for i in range(K.shape[0]):
    #     covariance_[i] = linalg.pinvh(K[i])

    return_list = [K, covariance_]
    if return_history:
        return_list.append(checks)
    if return_n_iter:
        return_list.append(iteration_ + 1)
    if return_n_linesearch:
        return_list.append(n_linesearch)
    return return_list


class TimeGraphLassoForwardBackward(TimeGraphLasso):
    """Sparse inverse covariance estimation with an l1-penalized estimator.

    Parameters
    ----------
    alpha : positive float, default 0.01
        Regularization parameter for precision matrix. The higher alpha,
        the more regularization, the sparser the inverse covariance.

    beta : positive float, default 1
        Regularization parameter to constrain precision matrices in time.
        The higher beta, the more regularization,
        and consecutive precision matrices in time are more similar.

    psi : {'laplacian', 'l1', 'l2', 'linf', 'node'}, default 'laplacian'
        Type of norm to enforce for consecutive precision matrices in time.

    rho : positive float, default 1
        Augmented Lagrangian parameter.

    over_relax : positive float, deafult 1
        Over-relaxation parameter (typically between 1.0 and 1.8).

    tol : positive float, default 1e-4
        Absolute tolerance to declare convergence.

    rtol : positive float, default 1e-4
        Relative tolerance to declare convergence.

    max_iter : integer, default 100
        The maximum number of iterations.

    verbose : boolean, default False
        If verbose is True, the objective function, rnorm and snorm are
        printed at each iteration.

    assume_centered : boolean, default False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    time_on_axis : {'first', 'last'}, default 'first'
        If data have time as the last dimension, set this to 'last'.
        Useful to use scikit-learn functions as train_test_split.

    update_rho_options : dict, default None
        Options for the update of rho. See `update_rho` function for details.

    compute_objective : boolean, default True
        Choose if compute the objective function during iterations
        (only useful if `verbose=True`).

    mode : {'admm'}, default 'admm'
        Minimisation algorithm. At the moment, only 'admm' is available,
        so this is ignored.

    Attributes
    ----------
    covariance_ : array-like, shape (n_times, n_features, n_features)
        Estimated covariance matrix

    precision_ : array-like, shape (n_times, n_features, n_features)
        Estimated pseudo inverse matrix.

    n_iter_ : int
        Number of iterations run.

    """

    def __init__(self, alpha=0.01, beta=1., time_on_axis='first', tol=1e-4,
                 max_iter=100, verbose=False, assume_centered=False,
                 compute_objective=True, eps=0.5, choose='gamma', lamda=1,
                 delta=1e-4, gamma=1., lamda_criterion='b', time_norm=1,
                 return_history=False, debug=False,
                 return_n_linesearch=False,
                 vareps=1e-5, stop_at=None, stop_when=1e-4):
        super(TimeGraphLassoForwardBackward, self).__init__(
            alpha=alpha, tol=tol, max_iter=max_iter,
            verbose=verbose, assume_centered=assume_centered,
            compute_objective=compute_objective, beta=beta,
            time_on_axis=time_on_axis)
        self.delta = delta
        self.gamma = gamma
        self.lamda_criterion = lamda_criterion
        self.time_norm = time_norm
        self.eps = eps
        self.choose = choose
        self.lamda = lamda
        self.return_history = return_history
        self.debug = debug
        self.return_n_linesearch = return_n_linesearch
        self.vareps = vareps
        self.stop_at = stop_at
        self.stop_when = stop_when

    def _fit(self, emp_cov, n_samples):
        """Fit the TimeGraphLasso model to X.

        Parameters
        ----------
        emp_cov : ndarray, shape (n_time, n_features, n_features)
            Empirical covariance of data.

        """
        if self.alpha == 'max':
            # use sklearn alpha max
            self.alpha = self.alpha_max(emp_cov, is_covariance=True)

        out = time_graph_lasso(
            emp_cov, n_samples=n_samples, alpha=self.alpha, beta=self.beta,
            tol=self.tol, max_iter=self.max_iter, verbose=self.verbose,
            return_n_iter=True, return_history=self.return_history,
            compute_objective=self.compute_objective,
            time_norm=self.time_norm, lamda_criterion=self.lamda_criterion,
            gamma=self.gamma, delta=self.delta, eps=self.eps,
            choose=self.choose, lamda=self.lamda, debug=self.debug,
            return_n_linesearch=self.return_n_linesearch,
            vareps=self.vareps,
            stop_at=self.stop_at, stop_when=self.stop_when)

        if self.return_history:
            if self.return_n_linesearch:
                self.precision_, self.covariance_, self.history_, self.n_iter_, self.n_linesearch_ = out
            else:
                self.precision_, self.covariance_, self.history_, self.n_iter_ = out
        else:
            if self.return_n_linesearch:
                self.precision_, self.covariance_, self.n_iter_, self.n_linesearch_ = out
            else:
                self.precision_, self.covariance_, self.n_iter_ = out
        return self

    def alpha_max(self, X, is_covariance=False):
        """Compute the alpha_max for the problem at hand, based on sklearn."""
        from sklearn.covariance.graph_lasso_ import alpha_max
        if is_covariance:
            emp_cov = X
        else:
            emp_cov = np.array([empirical_covariance(
                x, assume_centered=self.assume_centered) for x in X])
        return max(alpha_max(e) for e in emp_cov)


def get_lipschitz(data):
    """Get the Lipschitz constant for a specific loss function.

    Only square loss implemented.

    Parameters
    ----------
    data : (n, d) float ndarray
        data matrix
    loss : string
        the selected loss function in {'square', 'logit'}
    Returns
    ----------
    L : float
        the Lipschitz constant
    """
    n, p = data.shape

    if p > n:
        tmp = np.dot(data, data.T)
    else:
        tmp = np.dot(data.T, data)
    return np.linalg.norm(tmp, 2)



"""Utils for REGAIN package."""
from __future__ import division

import collections
import functools
import logging
import os
import sys
import warnings
# from collections import namedtuple
from contextlib import contextmanager

import numpy as np
import six
from numpy.linalg.linalg import LinAlgError
from scipy import stats
from scipy.spatial.distance import squareform
from six.moves import cPickle as pkl
from sklearn.metrics import average_precision_score


def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = collections.namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None, ) * len(T._fields)
    if isinstance(default_values, collections.Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T


convergence = namedtuple_with_defaults(
    'convergence', 'obj rnorm snorm e_pri e_dual precision')


@contextmanager
def suppress_stdout():
    """Suppress function output.

    Usage
    -----
    with suppress_stdout():
        function_with_outputs()
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def _ensure_filename_ending(filename, possible_extensions='.txt'):
    if isinstance(possible_extensions, six.string_types):
        possible_extensions = [possible_extensions]

    return filename + (
        '' if any(filename.endswith(end)
                  for end in possible_extensions) else possible_extensions[0])


def init_logger(filename, verbose=True):
    """Initialise logger."""
    logfile = _ensure_filename_ending(filename, ['.log', '.txt'])
    logging.shutdown()
    root_logger = logging.getLogger()
    for _ in list(root_logger.handlers):
        root_logger.removeHandler(_)
        _.flush()
        _.close()
    for _ in list(root_logger.filters):
        root_logger.removeFilter(_)
        _.flush()
        _.close()

    logging.basicConfig(
        filename=logfile, level=logging.INFO, filemode='w',
        format='%(levelname)s (%(asctime)-15s): %(message)s')
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO if verbose else logging.ERROR)
    stream_handler.setFormatter(
        logging.Formatter('%(levelname)s (%(asctime)-15s): %(message)s'))

    root_logger.addHandler(stream_handler)
    return logfile


def save_pickle(obj, filename):
    """Save pickle utility."""
    filename = _ensure_filename_ending(filename, '.pkl')
    with open(filename, 'wb') as f:
        pkl.dump(obj, f)


def load_pickle(filename):
    """Read pickle utility."""
    with open(filename, 'rb') as f:
        res = pkl.load(f)
    return res


def write_network(dataframe, filename):
    """Write a network as a list of interactions."""
    dataframe.stack().to_csv(filename)


def read_network(
        filename, threshold=1., full_network=True, fill_diagonal=True,
        delimiter="auto"):
    """Read a network from a list of interactions.

    Parameters
    ----------
    filename : str
        Filename to read from.
    threshold : float, optional
        Only get the top threshold edges.
    full_network : boolean, optional
        Choose if the network is written in full or only the upper triangular.
    fill_diagonal : boolean, optional
        Fill diagonal with the sum of absolute values in the row (for the
        positive definite constraint).
    """
    import pandas as pd
    if filename.endswith(".tsv") or filename.endswith(".tab"):
        if delimiter == 'auto':
            delimiter = '\t'
        elif delimiter != '\t':
            warnings.warn(
                "The extension is suggesting the filename is tab-"
                "separated. Please check you are using the correct "
                "separator.")
    elif filename.endswith(".csv"):
        if delimiter == 'auto':
            delimiter = ','
        elif delimiter != ',':
            warnings.warn(
                "The extension is suggesting the filename is comma-"
                "separated. Please check you are using the correct "
                "separator.")
    else:
        if delimiter == 'auto':
            raise ValueError("Unrecognized format. Please specify a separator")

    nn = pd.read_csv(filename, delimiter=delimiter, header=None)
    # the following suppose genes are in the form G1, ... G10
    columns = sorted(nn[0].unique(), key=lambda x: int(x[1:]))
    n_top_edges = int(nn.shape[0] * threshold / (2. if full_network else 1))

    nn = nn.sort_values(
        2, ascending=False)[:(2 if full_network else 1) * n_top_edges]

    net_julia = pd.DataFrame(columns=columns, index=columns,
                             dtype=float).fillna(0)

    for row in nn.itertuples():
        if row[3] > 0:
            net_julia.loc[row[1], row[2]] = \
                net_julia.loc[row[2], row[1]] = row[3]

    if fill_diagonal:
        np.fill_diagonal(net_julia.values, net_julia.sum(axis=1).values)
    return net_julia


def flatten(lst):
    """Flatten a list."""
    return [y for l in lst for y in flatten(l)] \
        if isinstance(lst, (list, np.ndarray)) else [lst]


def upper_to_full(a):
    """Convert the upper part to a full symmetric matrix."""
    n = int((np.sqrt(1 + 8 * a.shape[0]) - 1) / 2)
    A = np.zeros((n, n))
    idx = np.triu_indices(n)
    A[idx] = A[idx[::-1]] = a
    return A


def compose(*functions):
    """Compose two or more functions."""

    def compose2(f, g):
        return lambda x: f(g(x))

    return functools.reduce(compose2, functions, lambda x: x)


def error_rank(ells_true, ells_pred):
    """Compute the mean absolute error in rank between two matrices.

    Parameters
    ----------
    ells_{true, pred} : array-like, 3 dimensional
        Latent variable matrices for which to compare the rank.

    """
    ranks_true = np.array([np.linalg.matrix_rank(l) for l in ells_true])
    ranks_pred = np.array([np.linalg.matrix_rank(l) for l in ells_pred])
    return np.mean(np.abs(ranks_true - ranks_pred))


def normalize_matrix(x):
    """Normalize a matrix so to have 1 on the diagonal, in-place."""
    d = np.diag(x).reshape(1, x.shape[0])
    d = 1. / np.sqrt(d)
    x *= d
    x *= d.T


def error_norm(
        cov, comp_cov, norm='frobenius', scaling=True, squared=True,
        upper_triangular=False, nonzero=False, n=False):
    """Mean Squared Error between two covariance estimators.

    Parameters
    ----------
    cov, comp_cov : array-like, shape = [n_features, n_features]
        The covariance to compare with.

    norm : str
        The type of norm used to compute the error. Available error types:
        - 'frobenius' (default): sqrt(tr(A^t.A))
        - 'spectral': sqrt(max(eigenvalues(A^t.A))
        where A is the error ``(comp_cov - self.covariance_)``.

    scaling : bool
        If True (default), the squared error norm is divided by n_features.
        If False, the squared error norm is not rescaled.

    squared : bool
        Whether to compute the squared error norm or the error norm.
        If True (default), the squared error norm is returned.
        If False, the error norm is returned.

    Returns
    -------
    The Mean Squared Error (in the sense of the Frobenius norm) between
    `self` and `comp_cov` covariance estimators.

    """
    if n:
        comp_cov = comp_cov.copy()
        # / comp_cov.max()
        cov = cov.copy()
        # / cov.max()
        [normalize_matrix(c) for c in comp_cov]
        [normalize_matrix(c) for c in cov]

    # compute the error
    if upper_triangular:
        comp_cov = np.triu(comp_cov, 1)
        cov = np.triu(cov, 1)
    if nonzero:
        comp_cov = comp_cov[np.where(cov == 0)]
        cov = cov[np.where(cov == 0)]

    error = comp_cov - cov
    # compute the error norm
    if norm == "frobenius":
        squared_norm = np.sum(error**2)
    elif norm == "spectral":
        squared_norm = np.amax(np.linalg.svdvals(np.dot(error.T, error)))
    else:
        raise NotImplementedError(
            "Only spectral and frobenius norms are implemented")
    # optionally scale the error norm
    if scaling:
        scaling_factor = error.shape[0] if len(error.shape) < 3 \
            else np.prod(error.shape[:2]) * ((error.shape[1] - 1) / 2. if upper_triangular else error.shape[1])
        squared_norm = squared_norm / scaling_factor
    # finally get either the squared norm or the norm
    if squared:
        result = squared_norm
    else:
        result = np.sqrt(squared_norm)
    return result


def error_norm_time(
        cov, comp_cov, norm='frobenius', scaling=True, squared=True):
    """Mean Squared Error between two covariance estimators.

    Parameters
    ----------
    cov, comp_cov : array-like, shape = [n_time, n_features, n_features]
        The covariance to compare with.

    norm : str
        The type of norm used to compute the error. Available error types:
        - 'frobenius' (default): sqrt(tr(A^t.A))
        - 'spectral': sqrt(max(eigenvalues(A^t.A))
        where A is the error ``(comp_cov - self.covariance_)``.

    scaling : bool
        If True (default), the squared error norm is divided by n_features.
        If False, the squared error norm is not rescaled.

    squared : bool
        Whether to compute the squared error norm or the error norm.
        If True (default), the squared error norm is returned.
        If False, the error norm is returned.

    Returns
    -------
    The Mean Squared Error (in the sense of the Frobenius norm) between
    `self` and `comp_cov` covariance estimators.

    """
    return np.mean(
        [
            error_norm(x, y, norm=norm, scaling=scaling, squared=squared)
            for x, y in zip(cov, comp_cov)
        ])


def structure_error(
        true, pred, thresholding=False, eps=1e-2, no_diagonal=False):
    """Error in structure between a precision matrix and predicted.

    Parameters
    ----------
    true: array-like
        True matrix. In grpahical inference, if an entry is different from 0
        it is consider as an edge (inverse covariance).

    pred: array-like, shape=(d,d)
        Predicted matrix. In graphical inference, if an entry is different
        from 0 it is consider as an edge (inverse covariance).

    thresholding: bool, default False,
       Apply a threshold (with eps) to the `pred` matrix.

    eps : float, default = 1e-2
        Apply a threshold (with eps) to the `pred` matrix.

    """
    # avoid inplace modifications
    true = true.copy()
    pred = pred.copy()

    if true.ndim > 2:
        y_true = np.array(flatten([squareform(x, checks=None) for x in true]))
        y_pred = np.array(flatten([squareform(x, checks=None) for x in pred]))
    else:
        y_true = squareform(true, checks=None)
        y_pred = squareform(pred, checks=None)

    average_precision = average_precision_score(y_true > 0, y_pred)

    if thresholding:
        pred[np.abs(pred) < eps] = 0
    tn_to_remove = 0
    if no_diagonal:
        if true.ndim > 2:
            true = np.array([t - np.diag(np.diag(t)) for t in true])
            pred = np.array([t - np.diag(np.diag(t)) for t in pred])
            tn_to_remove = np.prod(true.shape[:2])
        else:
            true -= np.diag(np.diag(true))
            pred -= np.diag(np.diag(pred))
            tn_to_remove = true.shape[0]
    true[true != 0] = 1
    pred[pred != 0] = 2
    res = true + pred
    # from collections import Counter
    # c = Counter(res.flat)
    # tn, fn, fp, tp = c[0], c[1], c[2], c[3]
    TN = np.count_nonzero((res == 0).astype(float)) - tn_to_remove
    FN = np.count_nonzero((res == 1).astype(float))
    FP = np.count_nonzero((res == 2).astype(float))
    TP = np.count_nonzero((res == 3).astype(float))

    precision = TP / (TP + FP) if TP + FP > 0 else 0
    recall = TP / (TP + FN)
    miss_rate = FN / (TP + FN) or 1 - recall
    f1 = 2 * precision * recall / (precision + recall) \
        if precision + recall > 0 else 0

    accuracy = (TP + TN) / true.size
    prevalence = (TP + FN) / true.size

    fall_out = FP / (FP + TN) if (FP + TN) > 0 else 1
    specificity = TN / (FP + TN) if (FP + TN) > 0 else 1. - fall_out

    balanced_accuracy = 0.5 * (recall + specificity)
    false_discovery_rate = FP / (TP + FP) if TP + FP > 0 else 1 - precision
    false_omission_rate = FN / (FN + TN) if FN + TN > 0 else 0
    negative_predicted_value = TN / (FN + TN) if FN + TN > 0 else \
        1 - false_omission_rate

    positive_likelihood_ratio = recall / fall_out if fall_out > 0 else 0
    negative_likelihood_ratio = miss_rate / specificity \
        if specificity > 0 else 0
    diagnostic_odds_ratio = \
        positive_likelihood_ratio / negative_likelihood_ratio if \
        negative_likelihood_ratio > 0 else 0

    dictionary = dict(
        tp=TP, tn=TN, fp=FP, fn=FN, precision=precision, recall=recall, f1=f1,
        accuracy=accuracy, false_omission_rate=false_omission_rate,
        fdr=false_discovery_rate, npv=negative_predicted_value,
        prevalence=prevalence, miss_rate=miss_rate, fall_out=fall_out,
        specificity=specificity, plr=positive_likelihood_ratio,
        nlr=negative_likelihood_ratio, dor=diagnostic_odds_ratio,
        balanced_accuracy=balanced_accuracy,
        average_precision=average_precision)
    return dictionary


def is_pos_semidef(x, tol=1e-15):
    """Check if x is positive semi-definite."""
    eigs = np.linalg.eigvalsh(x)
    eigs[np.abs(eigs) < tol] = 0
    return np.all(eigs >= 0)


def is_pos_def(x, tol=1e-15, chol=True):
    """Check if x is positive definite."""
    if chol:
        try:
            np.linalg.cholesky(x)
            return True
        except LinAlgError:
            return False

    eigs = np.linalg.eigvalsh(x)
    eigs[np.abs(eigs) < tol] = 0
    return np.all(eigs > 0)


def positive_definite(x, tol=1e-15):
    if x.ndim == 2:
        return is_pos_def(x)
    return all(is_pos_def(y) for y in x)


def ensure_posdef(X, inplace=True):
    def _ensure_posdef_2d(X, inplace=True):
        if not inplace:
            raise NotImplementedError("Only inplace implemented")
        X.flat[::X.shape[0] +
               1] = np.abs(X - np.diag(np.diag(X))).sum(axis=1) + 0.1

    if X.ndim == 2:
        return _ensure_posdef_2d(X, inplace)
    for x in X:
        _ensure_posdef_2d(x, inplace)
    return


def threshold(a, threshmin=None, threshmax=None, newval=0):
    """Threshold. Replaces the deprecated scipy function."""
    a = np.ma.array(a, copy=True)
    mask = np.zeros(a.shape, dtype=bool)
    if threshmin is not None:
        mask |= (a < threshmin).filled(False)

    if threshmax is not None:
        mask |= (a > threshmax).filled(False)

    a[mask] = newval
    return a


def alpha_heuristic(emp_cov, n_samples, gamma=0.1):
    """An heuristic for GraphLasso alpha.

    XXX - need testing

    References
    ----------
    http://people.eecs.berkeley.edu/~elghaoui/Pubs/CvxTechCovSel_ICML.pdf
    """
    if emp_cov.ndim == 3:
        diag = np.diagonal(emp_cov, axis1=1, axis2=2)
        m = np.array([d[:, None].dot(d[None, :]) for d in diag]).max()
    elif emp_cov.ndim == 2:
        diag = np.diag(emp_cov)[:, None]
        m = np.max(diag.dot(diag.T))
    else:
        raise ValueError(emp_cov.ndim)

    t = stats.t.pdf(gamma, n_samples - 2) * 2
    num = t * m
    den = np.sqrt(n_samples - 2 + t * t)
    return num / den



"""Dataset generation module."""
from __future__ import division

import warnings
from functools import partial

import numpy as np
from scipy import signal
from scipy.spatial.distance import squareform
from sklearn.datasets.base import Bunch
from sklearn.utils import deprecated

from regain.utils import is_pos_def, is_pos_semidef, positive_definite


def normalize_matrix(x):
    """Normalize a matrix so to have 1 on the diagonal, in-place."""
    d = np.diag(x).reshape(1, x.shape[0])
    d = 1. / np.sqrt(d)
    x *= d
    x *= d.T


def make_dataset(n_samples=100, n_dim_obs=100, n_dim_lat=10, T=10,
                 mode=None, time_on_axis='first',
                 update_ell='l2', update_theta='l2',
                 normalize_starting_matrices=False, degree=2, epsilon=1e-2,
                 keep_sparsity=False, proportional=False, **kwargs):
    """Generate a synthetic dataset.

    Parameters
    ----------
    n_samples: int,
        number of samples to generate
    n_dim_obs: int,
        number of observed variables of the graph
    n_dim_lat: int,
        number of latent variables of the graph
    T: int,
        number of times
    mode: string,
        "evolving": generate a dataset with evolving observed and latent
                    variables that have a small Frobenious norm between two
                    close time points
        "fixed": generate a dataset with evolving observed and fixed latent
                    variables that have a small Frobenious norm between two
                    close time points
        "l1": generate a dataset with evolving observed and latent variables
              that differs for a small l1 norm
        "l1l2": generate a dataset with evolving observed variables that
                differs for a small l1 norm and evolving latent variables
                that differs for a small l2 norm
        "sin": generate a dataset with fixed latent variables and evolving
                observed variables that are generated from sin functions.
        func : use the user-defined function to generate the dataset. Such
            function should return, in this order, "thetas, thetas_obs, ells".
            See the other functions for an example.
    *kwargs: other arguments related to each specific data generation mode

    """
    n_dim_obs = int(n_dim_obs)
    n_dim_lat = int(n_dim_lat)
    n_samples = int(n_samples)

    modes = dict(
        evolving=make_l2l2,
        fixed=make_l2,
        fixedl2=make_l2,
        fixedl1=make_l1,
        yuan=generate_dataset_yuan,
        l1l2=generate_dataset_l1l2,
        norm=make_l2l2_norm, l1l1=generate_dataset_l1l1,

        # the previous are deprecated
        sin=make_sin,
        fixed_sparsity=make_fixed_sparsity, sincos=make_sin_cos,
        fede=make_fede, sklearn=make_sparse_low_rank,
        ma=make_ma_xue_zou, mak=make_ma_xue_zou_rand_k)

    if mode is not None:
        # mode overrides other parameters, for back compatibility
        func = mode if callable(mode) else modes.get(mode, None)
        if func is None:
            raise ValueError("Unknown mode %s. "
                             "Choices are: %s" % (mode, modes.keys()))
        kwargs.update(degree=degree, epsilon=epsilon,
                      keep_sparsity=keep_sparsity, proportional=proportional)
    else:
        func = partial(
            make_covariance, update_ell=update_ell, update_theta=update_theta,
            normalize_starting_matrices=normalize_starting_matrices,
            degree=degree, epsilon=epsilon, keep_sparsity=keep_sparsity,
            proportional=proportional)

    thetas, thetas_obs, ells = func(n_dim_obs, n_dim_lat, T, **kwargs)
    sigmas = map(np.linalg.inv, thetas_obs)
    # map(normalize_matrix, sigmas)  # in place

    data = np.array([np.random.multivariate_normal(
        np.zeros(n_dim_obs), sigma, size=n_samples) for sigma in sigmas])

    if time_on_axis == "last":
        data = data.transpose(1, 2, 0)
    return Bunch(data=data, thetas=np.array(thetas),
                 thetas_observed=np.array(thetas_obs),
                 ells=np.array(ells))


def make_ell(n_dim_obs=100, n_dim_lat=10):
    """Doc."""
    K_HO = np.zeros((n_dim_lat, n_dim_obs))
    for i in range(n_dim_lat):
        percentage = int(n_dim_obs * 0.8)
        indices = np.random.randint(0, high=n_dim_obs, size=percentage)
        K_HO[i, indices] = np.random.rand(percentage) * 0.12

    K_HO /= np.sum(K_HO, axis=1)[:, None] / 2
    L = K_HO.T.dot(K_HO)
    assert(is_pos_semidef(L))
    assert np.linalg.matrix_rank(L) == n_dim_lat
    # from sklearn.datasets import make_low_rank_matrix
    # L = make_low_rank_matrix(n_dim_obs, n_dim_obs, effective_rank=n_dim_lat)
    # L = (L + L.T) / 2.
    # print L
    return L, K_HO


def make_starting(n_dim_obs=100, n_dim_lat=10, degree=2, normalize=False):
    """Generate starting theta, theta_observed, L, K_HO."""
    L, K_HO = make_ell(n_dim_obs, n_dim_lat)

    if normalize:
        theta = np.zeros((n_dim_obs, n_dim_obs))
        for i in range(n_dim_obs):
            possible_idx = list(set(range(n_dim_obs)) - (
                set(np.nonzero(theta[i, :])[0]) |
                set(np.where(np.count_nonzero(theta, axis=1) > degree)[0])))
            if not possible_idx:
                continue
            n_choice = degree - (np.count_nonzero(theta[i, :]) - 1)
            if n_choice > 0:
                indices = np.random.choice(possible_idx, n_choice)
            theta[i, indices] = theta[indices, i] = 1. / degree

        theta.flat[::n_dim_obs+1] = np.sum(theta, axis=1) + 0.002

    else:
        theta = np.eye(n_dim_obs)
        for i in range(n_dim_obs):
            possible_idx = list(set(range(n_dim_obs)) - (
                set(np.nonzero(theta[i, :])[0]) |
                set(np.where(np.count_nonzero(theta, axis=1) > degree)[0])))
            if not possible_idx:
                continue
            n_choice = degree - (np.count_nonzero(theta[i, :]) - 1)
            if n_choice > 0:
                indices = np.random.choice(possible_idx, n_choice)
                theta[i, indices] = theta[indices, i] = .5 / degree

    assert(is_pos_def(theta))
    theta_observed = theta - L
    assert(is_pos_def(theta_observed))
    return theta, theta_observed, L, K_HO


def update_theta_l2(
        theta_old, n_dim_obs, degree, epsilon, keep_sparsity=False,
        indices=None):
    addition = np.zeros_like(theta_old)
    for i in range(n_dim_obs):
        if keep_sparsity:
            ii = indices[i]
        else:
            ii = np.random.randint(0, n_dim_obs, size=degree)
        addition[i, ii] = np.random.randn(len(ii))
    addition[np.triu_indices(n_dim_obs)[::-1]] = \
        addition[np.triu_indices(n_dim_obs)]
    addition *= epsilon / np.linalg.norm(addition)
    np.fill_diagonal(addition, 0)
    theta = theta_old + addition
    theta[np.abs(theta) < 2 * epsilon / n_dim_obs] = 0
    return theta


def update_theta_l1(theta_init, no, n_dim_obs):
    theta = theta_init.copy()
    rows = np.zeros(no)
    cols = np.zeros(no)
    while (np.any(rows == cols)):
        rows = np.random.randint(0, n_dim_obs, no)
        cols = np.random.randint(0, n_dim_obs, no)
    for r, c in zip(rows, cols):
        theta[r, c] = np.random.choice([0.12, 0, 0]) if theta[r, c] == 0 else .06  # np.random.rand(1) * .35
        theta[c, r] = theta[r, c]
    assert(is_pos_def(theta))
    return theta


def update_ell_l2(K_HO_old, epsilon, n_dim_obs):
    K_HO = K_HO_old.copy()
    addition = np.random.rand(*K_HO.shape)
    addition *= epsilon / np.linalg.norm(addition)
    K_HO += addition
    K_HO /= np.sum(K_HO, axis=1)[:, None] / 2.
    # K_HO *= 0.12
    K_HO[np.abs(K_HO) < epsilon / n_dim_obs] = 0
    return K_HO.T.dot(K_HO), K_HO


def make_covariance(
    n_dim_obs=100, n_dim_lat=10, T=10, update_ell='l1', update_theta='l1',
    normalize_starting_matrices=True, degree=2, epsilon=1e-2,
        keep_sparsity=False, proportional=False):
    no = int(np.ceil(n_dim_obs / 20)) if proportional else 1

    theta, theta_observed, L, K_HO = make_starting(
        n_dim_obs, n_dim_lat, degree, normalize=normalize_starting_matrices)

    thetas = [theta]
    thetas_obs = [theta_observed]
    ells = [L]
    K_HOs = [K_HO]

    idx = [np.nonzero(row)[0] for row in theta] if keep_sparsity else None
    for i in range(1, T):
        if update_theta == 'l1':
            if keep_sparsity:
                warnings.warn("keep_sparsity is specified but is not "
                              "implemented with l1.")
            theta = update_theta_l1(thetas[-1], no, n_dim_obs)
        elif update_theta == 'l2':
            theta = update_theta_l2(thetas[-1], n_dim_obs, degree, epsilon,
                                    keep_sparsity=keep_sparsity, indices=idx)
        else:
            raise ValueError(update_theta)

        if update_ell == 'l1':
            K_HO = K_HOs[-1].copy()
            picks = np.random.permutation(K_HO.size)[:no]
            K_HO = K_HO.ravel()
            for p in picks:
                K_HO[p] = np.random.choice([0.12, 0, 0]) if K_HO[p] == 0 else 0
            K_HO = np.reshape(K_HO, (n_dim_lat, n_dim_obs))
            L = K_HO.T.dot(K_HO)
        elif update_ell == 'l2':
            L, K_HO = update_ell_l2(K_HOs[-1], epsilon, n_dim_obs)
        elif update_ell == 'yuan':
            K_HO = K_HOs[-1].copy()
            c = np.random.randint(0, n_dim_obs, 1)
            r = np.random.randint(0, n_dim_lat, 1)
            K_HO[r, c] = 0.12 if K_HO[r, c] == 0 else 0
            L = K_HO.T.dot(K_HO)
        elif update_ell == 'fixed':
            pass  # L is fixed
        else:
            raise ValueError(update_ell)

        assert np.linalg.matrix_rank(L) == n_dim_lat
        assert(is_pos_semidef(L))
        assert(is_pos_def(theta - L))

        thetas.append(theta)
        thetas_obs.append(theta - L)
        ells.append(L)
        K_HOs.append(K_HO)

    return thetas, thetas_obs, ells


@deprecated("it will be removed in v0.2.0. Use `make_covariance` instead")
def generate_dataset_l1l1(
        n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Generate matrices according to a l1-l1 model."""
    return make_covariance(
        n_dim_lat=n_dim_lat, n_dim_obs=n_dim_obs, T=T, update_ell='l1',
        update_theta='l1', **kwargs)


@deprecated("it will be removed in v0.2.0. Use `make_covariance` instead")
def generate_dataset_l1l2(n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """DESCRIZIONE, PRIMA O POI."""
    return make_covariance(
        n_dim_lat=n_dim_lat, n_dim_obs=n_dim_obs, T=T, update_ell='l2',
        update_theta='l1', **kwargs)


@deprecated("it will be removed in v0.2.0. Use `make_covariance` instead")
def generate_dataset_yuan(n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Yuan (2012) model."""
    # number of links for each node
    return make_covariance(
        n_dim_lat=n_dim_lat, n_dim_obs=n_dim_obs, T=T, update_ell='yuan',
        **kwargs)


@deprecated("it will be removed in v0.2.0. Use `make_covariance` instead")
def make_l2l2(
        n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Generate dataset with evolving L."""
    return make_covariance(
        n_dim_lat=n_dim_lat, n_dim_obs=n_dim_obs, T=T, update_ell='l2',
        update_theta='l2', normalize_starting_matrices=False, **kwargs)


@deprecated("it will be removed in v0.2.0. Use `make_covariance` instead")
def make_l2l2_norm(n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Generate dataset with evolving L."""
    return make_covariance(
        n_dim_lat=n_dim_lat, n_dim_obs=n_dim_obs, T=T, update_ell='l2',
        update_theta='l2', normalize_starting_matrices=True, **kwargs)


@deprecated("it will be removed in v0.2.0. Use `make_covariance` instead")
def make_l2(n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Generate precisions with a fixed L matrix."""
    return make_covariance(
        n_dim_lat=n_dim_lat, n_dim_obs=n_dim_obs, T=T, update_ell='fixed',
        update_theta='l2', normalize_starting_matrices=False, **kwargs)


@deprecated("it will be removed in v0.2.0. Use `make_covariance` instead")
def make_l1(n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Generate precisions with a fixed L matrix."""
    return make_covariance(
        n_dim_lat=n_dim_lat, n_dim_obs=n_dim_obs, T=T, update_ell='fixed',
        update_theta='l1', normalize_starting_matrices=False, **kwargs)


def make_fixed_sparsity(n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Generate precisions with a fixed L matrix and sparsity."""
    degree = kwargs.get('degree', 2)
    epsilon = kwargs.get('epsilon', 1e-2)
    t = kwargs.get('changing_time', T / 2.)
    theta, theta_observed, L, K_HO = make_starting(
        n_dim_obs, n_dim_lat, degree, normalize=False)

    theta = np.abs(theta)
    theta_observed = theta - L
    nonzeros = np.nonzero(theta-np.diag(theta))

    thetas = [theta]
    thetas_obs = [theta_observed]

    for i in range(1, T):
        theta = thetas[-1].copy()
        if t < T:
            theta[nonzeros] += np.random.randn(nonzeros[0].size) * 0.1
        else:
            theta[nonzeros] -= np.random.randn(nonzeros[0].size) * 0.1
        theta = np.abs(theta)
        theta = (theta + theta.T) / 2.
        theta[theta < epsilon] = 0
        theta.flat[::n_dim_obs + 1] = np.sum(np.abs(theta), axis=1) \
                                    + np.sum(np.abs(L), axis=1) + .1
        nonzeros = np.nonzero(theta-np.diag(theta))

        theta_observed = theta - L
        assert is_pos_def(theta_observed)
        thetas.append(theta)
        thetas_obs.append(theta_observed)

    return thetas, thetas_obs, np.array([L] * T)


def make_sin(n_dim_obs, n_dim_lat, T, shape='smooth', closeness=1,
             normalize=False, **kwargs):
    upper_idx = np.triu_indices(n_dim_obs, 1)
    n_interactions = len(upper_idx[0])
    x = np.tile(np.linspace(0, (T-1.) / closeness, T), (n_interactions, 1))
    phase = np.random.rand(n_interactions, 1)
    freq = np.random.rand(n_interactions, 1) - .50
    A = (np.random.rand(n_interactions, 1) + 1) / 2.

    if shape == 'smooth':
        y = A * np.sin(2. * np.pi * freq * x + phase)
    else:
        y = A * signal.square(2 * np.pi * freq * x + phase, duty=.5)

    # threshold
    y = np.maximum(y, 0)

    Y = np.array([squareform(y[:, j]) + np.diag(np.sum(squareform(y[:, j]), axis=1))
                  for j in range(y.shape[1])])

    if normalize:
        map(normalize_matrix, Y)  # in place
    assert positive_definite(Y)

    return Y, Y, np.zeros_like(Y)


def make_sin_cos(n_dim_obs=100, n_dim_lat=10, T=10, **kwargs):
    """Variables follow sin and cos evolution. L is fixed."""
    degree = kwargs.get('degree', 2)
    eps = kwargs.get('epsilon', 1e-2)
    L, K_HO = make_ell(n_dim_obs, n_dim_lat)

    phase = np.random.randn(n_dim_obs, n_dim_obs) * np.pi
    upper_idx_diag = np.triu_indices(n_dim_obs)
    phase[upper_idx_diag[::-1]] = phase[upper_idx_diag]

    upper_idx = np.triu_indices(n_dim_obs, 1)
    clip = np.zeros((n_dim_obs, n_dim_obs))
    picks = np.random.permutation(len(upper_idx[0]))
    dim = int(len(upper_idx[0]) * degree)
    picks = picks[:dim]
    clip1 = clip[upper_idx].ravel()
    clip1[picks] = 1
    clip[upper_idx[::-1]] = clip[upper_idx] = clip1

    thetas = np.array([np.eye(n_dim_obs) for i in range(T)])

    x = np.linspace(0, 2*np.pi, T)
    for i in range(T):
        for r in range(n_dim_obs):
            for c in range(n_dim_obs):
                if r == c:
                    continue
                if clip[r, c]:
                    thetas[i, r, c] = np.sin((x[i]+phase[r, c])/T**2)
                else:
                    thetas[i, r, c] = np.sin((x[i]+phase[r, c]))
        thetas[i][clip == 1] = np.clip(thetas[i][clip == 1], 0, 1)
        thetas[i][np.abs(thetas[i]) < eps] = 0

        assert(is_pos_def(thetas[i]))
        theta_observed = thetas[i] - L
        assert(is_pos_def(theta_observed))
        thetas_obs = [theta_observed]

    return thetas, thetas_obs, np.array([L]*T)


def make_fede(
        n_dim_obs=3, n_dim_lat=2, T=10, epsilon=1e-3, n_samples=50, **kwargs):
    """Generate dataset (new version)."""
    b = np.random.rand(1, n_dim_obs)
    es, Q = np.linalg.eigh(b.T.dot(b))  # Q random

    b = np.random.rand(1, n_dim_obs)
    es, R = np.linalg.eigh(b.T.dot(b))  # R random

    start_sigma = np.random.rand(n_dim_obs) + 1
    start_lamda = np.zeros(n_dim_obs)
    idx = np.random.randint(n_dim_obs, size=n_dim_lat)
    start_lamda[idx] = np.random.rand(n_dim_lat)

    Ks = []
    Ls = []
    Kobs = []

    for i in range(T):
        K = np.linalg.multi_dot((Q, np.diag(start_sigma), Q.T))
        L = np.linalg.multi_dot((R, np.diag(start_lamda), R.T))

        K[np.abs(K) < epsilon] = 0  # enforce sparsity on K

        # assert is_pos_def(K - L)
        # assert is_pos_semidef(L)

        start_sigma += 1 + np.random.rand(n_dim_obs)
        start_lamda[idx] += np.random.rand(n_dim_lat) * 2 - 1
        start_lamda = np.maximum(start_lamda, 0)

        Ks.append(K)
        Ls.append(L)
        Kobs.append(K - L)

    return Ks, Kobs, Ls


def make_sparse_low_rank(
        n_dim_obs=3, n_dim_lat=2, T=10, epsilon=1e-3, **kwargs):
    """Generate dataset (new new version)."""
    from sklearn.datasets import make_sparse_spd_matrix, make_low_rank_matrix

    K = make_sparse_spd_matrix(n_dim_obs)
    L = make_low_rank_matrix(n_dim_obs, n_dim_obs, effective_rank=n_dim_lat)

    Ks = [K]
    Ls = [L]
    Kobs = [K - L]

    for i in range(1, T):
        K = K + make_sparse_spd_matrix(n_dim_obs)
        L = L + make_low_rank_matrix(n_dim_obs, n_dim_obs,
                                     effective_rank=n_dim_lat)

        # assert is_pos_def(K - L)
        # assert is_pos_semidef(L)

        Ks.append(K)
        Ls.append(L)
        Kobs.append(K - L)

    return Ks, Kobs, Ls


def make_ma_xue_zou(n_dim_obs=12, n_latent=3, T=1, epsilon=1e-3, sparsity=0.1):
    """Generate the dataset as in Ma, Xue, Zou (2012)."""
    # p = n_dim_obs + n_latent  # int(n_dim_obs * 0.05)
    p = n_dim_obs + int(n_dim_obs * 0.05)
    po = n_dim_obs
    ph = p - n_dim_obs
    W = np.zeros((p, p))
    non_zeros = int(round(p*p*sparsity))
    picks = np.random.permutation(p*p)[:non_zeros]
    W = W.ravel(order='F')
    W[picks] = np.random.randn(non_zeros)
    W = np.reshape(W, (p, p), order="F")

    C = W.T.dot(W)
    C[:po, po:] += 0.5 * np.random.randn(po, ph)
    C = (C + C.T) / 2.

    C = np.clip(C - np.diag(np.diag(C)), -1, 1)
    eig, Q = np.linalg.eigh(C)
    K = C + max(-1.2 * np.min(eig), 0.001) * np.eye(p)
    K_O = K[:po, :po]
    K_OH = K[:po, po:]
    K_HO = K[po:, :po]
    K_H = K[po:, po:]

    # L = np.divide(K_OH, K_H.dot(K_HO))
    assert np.allclose(K_OH, K_HO.T)
    L = np.linalg.multi_dot((K_OH, np.linalg.inv(K_H), K_HO))
    K_O_tilde = K_O - L
    assert is_pos_def(K_O_tilde)
    assert is_pos_semidef(K_H)
    assert np.linalg.matrix_rank(L) == ph
    # print(ph)

    N = 5 * po
    print("Note that, with this method, the n_samples should be %d" % N)
    return [K_O] * T, [K_O_tilde] * T, [L] * T


def make_ma_xue_zou_rand_k(
        n_dim_obs=12, n_latent=3, T=1, epsilon=1e-3, sparsity=0.1):
    """Generate the dataset as in Ma, Xue, Zou (2012)."""
    # p = n_dim_obs + n_latent  # int(n_dim_obs * 0.05)
    p = n_dim_obs + int(n_dim_obs * 0.05)
    po = n_dim_obs
    ph = p - n_dim_obs
    nnzr = int(sparsity * (np.triu_indices(p, 1)[0].size))

    # Generate A, the original inverse covariance, with random sparsity pattern...
    A = np.eye(p)
    idx = np.vstack(np.triu_indices(p, 1))
    idx = idx[:, np.random.choice(idx.shape[1], nnzr, replace=False)]
    idx = (idx[0], idx[1])
    A[idx] = np.sign(np.random.rand(nnzr) - .5)
    A[np.triu_indices(p, 1)[::-1]] = A[np.triu_indices(p, 1)]

    # A is the gound truth inverse covariance matrix
    K = A.dot(A.T) + 1e-6 * np.eye(p)
    K = A
    K_O = K[:po, :po]
    K_OH = K[:po, po:]
    K_HO = K[po:, :po]
    K_H = K[po:, po:]
    L = np.linalg.multi_dot((K_OH, np.linalg.inv(K_H), K_HO))
    K_O_tilde = K_O - L

    N = 5 * po
    print("Note that, with this method, the n_samples should be %d" % N)
    return [K_O] * T, [K_O_tilde] * T, [L] * T



import numpy as np


def sample(K, nu=1, p=3):
    """Sample from a GP.

    Parameters
    ----------
    K : ndarray, shape (n, n)
        Temporal kernel between n time points.
    nu : type
        Number of replicates.
    p : type
        Number of time series to generate.

    Returns
    -------
    u : ndarray, shape (nu, p, K.shape[0])
        Samples from a Gaussian process.

    """
    u = np.empty((nu, p, K.shape[0]))
    for j in range(nu):
        u[j] = np.random.multivariate_normal(np.zeros(K.shape[0]), K, size=p)
    return u



"""Useful proximal functions."""
import warnings
from functools import partial

import collections

import numpy as np
from six.moves import range, zip
from sklearn.utils.extmath import squared_norm

from regain.update_rules import update_rho
from regain.utils import convergence

try:
    from prox_tv import tv1_1d, tvp_1d, tvgen, tvp_2d
except:
    # fused lasso prox cannot be used
    pass


def soft_thresholding(a, lamda):
    """Soft-thresholding."""
    return np.sign(a) * np.maximum(np.abs(a) - lamda, 0)


def _soft_thresholding_od_2d(a, lamda):
    # this assumes array is 2-dimensional
    # no check is performed for optimisation
    soft = soft_thresholding(a, lamda)
    np.fill_diagonal(soft, np.diag(a))
    return soft


def soft_thresholding_od(a, lamda):
    """Off-diagonal soft-thresholding."""
    if a.ndim > 2:
        out = np.empty_like(a)
        if not isinstance(lamda, collections.Iterable):
            lamda = np.repeat(lamda, a.shape[0])
        else:
            assert lamda.shape[0] == a.shape[0]

        for t in range(a.shape[0]):
            out[t] = _soft_thresholding_od_2d(a[t], lamda[t])
    else:
        out = _soft_thresholding_od_2d(a, lamda)
    return out


def soft_thresholding_vector(a, lamda):
    """Soft-thresholding for vectors."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return np.maximum(1 - lamda / np.linalg.norm(a), 0) * a


def _blockwise_soft_thresholding_2d(a, lamda):
    """Proximal operator for l2 norm, a is two-dimensional."""
    return np.array([soft_thresholding_vector(aa, lamda) for aa in a.T]).T


def blockwise_soft_thresholding(a, lamda):
    """Proximal operator for l2 norm."""
    if a.ndim > 2:
        out = np.empty_like(a, dtype=float)
        for t in range(a.shape[0]):
            out[t] = _blockwise_soft_thresholding_2d(a[t], lamda)
    else:
        out = _blockwise_soft_thresholding_2d(a, lamda)
    return out


def blockwise_soft_thresholding_symmetric(a, lamda):
    """Proximal operator for l2 norm, for symmetric matrices (last 2 axes)."""
    col_norms = np.linalg.norm(a, axis=1)
    ones_vect = np.ones(a.shape[1])

    if a.ndim == 2:
        return np.dot(
            a, np.diag((ones_vect - lamda / col_norms) * (col_norms > lamda)))

    output = np.empty_like(a, dtype=float)
    for i, (x, c_norm) in enumerate(zip(a, col_norms)):
        output[i] = np.dot(
            x, np.diag((ones_vect - lamda / c_norm) * (c_norm > lamda)))
    return output


# %% Perform prox operator:   min_x (1/2t)||x-w||^2 subject to |x|<=radius
# function [ z ] = project_1ball( z,radius )
# %  By Moreau's identity, projection onto 1-norm ball can be computed
# %  using the proximal of the conjugate problem, which is L-infinity
# %  minimization.
#   z = z -  prox_infinityNorm(z,radius);
# end
# %% Perform prox operator:   min ||x||_inf + (1/2t)||x-w||^2
# function [ xk ] = prox_infinityNorm( w,t )
#     N = length(w);
#     wabs = abs(w);
#     ws = (cumsum(sort(wabs,'descend'))- t)./(1:N)';
#     alphaopt = max(ws);
#     if alphaopt>0
#       xk = min(wabs,alphaopt).*sign(w); % truncation step
#     else
#       xk = zeros(size(w)); % if t is big, then solution is zero
#     end
# end
def prox_linf_1d(a, lamda):
    """Proximal operator for the l-inf norm.

    Since there is no closed-form, we can minimize it with scipy.
    """
    from scipy.optimize import minimize

    def _f(x):
        return lamda * np.linalg.norm(x, np.inf) + \
            .5 * np.power(np.linalg.norm(a - x), 2)

    return minimize(_f, a).x


def prox_linf(a, lamda):
    """Proximal operator for l-inf norm."""
    x = np.zeros_like(a)
    for t in range(a.shape[0]):
        x[t] = np.array(
            [prox_linf_1d(a[t, :, j], lamda) for j in range(a.shape[1])]).T
    return x


def prox_logdet(a, lamda):
    """Time-varying latent variable graphical lasso prox."""
    es, Q = np.linalg.eigh(a)
    xi = (-es + np.sqrt(np.square(es) + 4. / lamda)) * lamda / 2.
    return np.linalg.multi_dot((Q, np.diag(xi), Q.T))


def prox_logdet_ala_ma(a, lamda):
    es, Q = np.linalg.eigh(a)
    xi = (-es + np.sqrt(np.square(es) + 4. * lamda)) / 2.
    return np.linalg.multi_dot((Q, np.diag(xi), Q.T))


def prox_trace_indicator(a, lamda):
    """Time-varying latent variable graphical lasso prox."""
    es, Q = np.linalg.eigh(a)
    xi = np.maximum(es - lamda, 0)
    return np.linalg.multi_dot((Q, np.diag(xi), Q.T))


def prox_laplacian(a, lamda):
    """Prox for l_2 square norm, Laplacian regularisation."""
    return a / (1 + 2. * lamda)


def prox_node_penalty(A_12, lamda, rho=1, tol=1e-4, rtol=1e-2, max_iter=500):
    """Lamda = beta / (2. * rho).

    A_12 = np.vstack((A_1, A_2))
    """
    n_time, _, n_dim = A_12.shape

    U_1 = np.full((A_12.shape[0], n_dim, n_dim), 1. / n_dim, dtype=float)
    U_2 = np.copy(U_1)
    Y_1 = np.copy(U_1)
    Y_2 = np.copy(U_1)

    C = np.hstack((np.eye(n_dim), -np.eye(n_dim), np.eye(n_dim)))
    inverse = np.linalg.inv(C.T.dot(C) + 2 * np.eye(3 * n_dim))

    V = np.zeros_like(U_1)
    W = np.zeros_like(U_1)
    V_old = np.zeros_like(U_1)
    W_old = np.zeros_like(U_1)

    for iteration_ in range(max_iter):
        A = (
            Y_1 - Y_2 - W - U_1 +
            (W.transpose(0, 2, 1) - U_2).transpose(0, 2, 1)) / 2.
        V = blockwise_soft_thresholding_symmetric(A, lamda=lamda)

        A = np.concatenate(((V + U_2).transpose(0, 2, 1), A_12), axis=1)
        D = V + U_1
        # Z = np.linalg.solve(C.T*C + eta*np.identity(3*n), - C.T*D + eta* A)
        Z = np.empty_like(A)
        for i, (A_i, D_i) in enumerate(zip(A, D)):
            Z[i] = inverse.dot(2 * A_i - C.T.dot(D_i))
        W, Y_1, Y_2 = (Z[:, i * n_dim:(i + 1) * n_dim, :] for i in range(3))

        # update residuals
        delta_U_1 = V + W - (Y_1 - Y_2)
        delta_U_2 = V - W.transpose(0, 2, 1)
        U_1 += delta_U_1
        U_2 += delta_U_2

        # diagnostics
        rnorm = np.sqrt(squared_norm(delta_U_1) + squared_norm(delta_U_2))
        snorm = rho * np.sqrt(
            squared_norm(W - W_old) + squared_norm(V + W - V_old - W_old))
        check = convergence(
            obj=np.nan, rnorm=rnorm, snorm=snorm,
            e_pri=np.sqrt(2 * V.size) * tol + rtol * max(
                np.sqrt(squared_norm(W) + squared_norm(V + W)),
                np.sqrt(squared_norm(V) + squared_norm(Y_1 - Y_2))),
            e_dual=np.sqrt(2 * V.size) * tol +
            rtol * rho * np.sqrt(squared_norm(U_1) + squared_norm(U_2)))
        W_old = W.copy()
        V_old = V.copy()

        # if np.linalg.norm(delta_U_1, 'fro') < tol and \
        #         np.linalg.norm(delta_U_2, 'fro') < tol:
        if check.rnorm <= check.e_pri and check.snorm <= check.e_dual:
            break
        rho_new = update_rho(rho, rnorm, snorm, iteration=iteration_)
        # scaled dual variables should be also rescaled
        U_1 *= rho / rho_new
        U_2 *= rho / rho_new
        rho = rho_new
    else:
        warnings.warn("Node norm did not converge.")
    return Y_1, Y_2


def prox_FL(
        a, beta, lamda, p=1, symmetric=False, use_matlab=False, optimize=True):
    """Fused Lasso prox.

    It is calculated as the Total variation prox + soft thresholding
    on the solution, as in
    http://ieeexplore.ieee.org/abstract/document/6579659/
    """
    # if any([any(np.diag(x) < 0) for x in a]):
    #     for a_i in a:
    #         np.fill_diagonal(a_i, np.sum(np.abs(a_i), axis=1))
    if optimize:
        Y = tvgen(a, [beta], [1], [p], n_threads=32, max_iters=30)

    else:
        Y = np.empty_like(a)
        # if use_matlab:
        #     from regain.wrapper.tv_condat import wrapper
        #     func = wrapper.total_variation_condat
        # else:
        func = tv1_1d if p == 1 else partial(tvp_1d, p=p)

        if symmetric:
            x, y = np.triu_indices_from(a[0])
            b = np.vstack(a.transpose(1, 2, 0))
            upper_ind = x * a.shape[1] + y
            Z = np.zeros_like(b)
            Z[upper_ind] = [func(row, beta) for row in b[upper_ind]]

            e = np.array(np.split(Z, a.shape[1], axis=0)).transpose(2, 0, 1)
            Y = (e + e.transpose(0, 2, 1)) / (
                np.array([np.eye(a.shape[1]) for i in range(a.shape[0])]) + 1)
        else:
            for i in range(np.power(a.shape[1], 2)):
                solution = func(a.flat[i::np.power(a.shape[1], 2)], beta)
                Y.flat[i::np.power(a.shape[1], 2)] = solution

    # fused-lasso (soft-thresholding on the solution)
    Y_soft = soft_thresholding(Y, lamda)

    # restore diagonal
    for y_s, y_fv in zip(Y_soft, Y):
        np.fill_diagonal(y_s, np.diag(y_fv))
    return Y_soft

    # Y = np.empty_like(a)
    # for i in range(a.shape[1]):
    #     for j in range(i, a.shape[2]):
    #         solution = tv1_1d(a[:, i, j], beta)
    #         # fused-lasso (soft-thresholding on the solution)
    #         if i != j:
    #             solution = soft_thresholding_sign(solution, lamda)
    #         Y[:, i, j] = solution
    # return Y

    # not work
    # x = np.vstack(a.transpose((2, 1, 0)))  # each row is the time evolution
    # assert np.allclose(np.array(xx), x)
    # x = np.apply_along_axis(
    #     compose(partial(soft_thresholding_sign, lamda=lamda),
    #             partial(tv1_1d, w=beta)), 1, x)
    # # return np.array(np.vsplit(x, x.shape[1])).transpose((2, 1, 0))
    # Z = np.array(np.vsplit(x, a.shape[1]))#.transpose((2, 1, 0))



"""Solve group lasso problem via ADMM.

More information can be found in the paper linked at:
http://www.stanford.edu/~boyd/papers/distr_opt_stat_learning_admm.html
"""
import numpy as np
from six.moves import range

from regain.linear_model.lasso_ import lu_factor
from regain.prox import soft_thresholding
from regain.utils import flatten


def group_lasso(A, b, lamda=1.0, groups=None, rho=1.0, alpha=1.0, max_iter=1000,
                tol=1e-4, rtol=1e-2, return_history=False):
    r"""Group Lasso solver.

    Solves the following problem via ADMM
       minimize 1/2*|| Ax - b ||_2^2 + \lambda sum(norm(x_i))

    The input p is a K-element vector giving the block sizes n_i, so that x_i
    is in R^{n_i}.

    Parameters
    ----------
    A : array-like, 2-dimensional
        Input matrix.
    b : array-like, 1-dimensional
        Output vector.
    lamda : float, optional
        Regularisation parameter.
    groups : list
        Groups of variables.
    rho : float, optional
        Augmented Lagrangian parameter.
    alpha : float, optional
        Over-relaxation parameter (typically between 1.0 and 1.8).
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.

    Returns
    -------
    x : numpy.array
        Solution to the problem.
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    n_samples, n_features = A.shape

    # check valid partition
    if not np.allclose(flatten(groups), np.arange(n_features)):
        raise ValueError("Invalid partition in groups. "
                         "Groups must be non-overlapping and each variables "
                         "must be selected")

    # % save a matrix-vector multiply
    Atb = A.T.dot(b)

    # ADMM solver
    x = np.zeros(n_features)
    z = np.zeros(n_features)
    u = np.zeros(n_features)

    # % pre-factor
    L, U = lu_factor(A, rho)

    hist = []
    for _ in range(max_iter):
        # % x-update
        q = Atb + rho * (z - u)  # % temporary value
        if n_samples >= n_features:
            x = np.linalg.lstsq(U, np.linalg.lstsq(L, q)[0])[0]
        else:
            x = q - A.T.dot(
                np.linalg.lstsq(
                    U, np.linalg.lstsq(
                        L, A.dot(q))[0])[0]) / rho
            x /= rho

        # % z-update with relaxation
        zold = z
        x_hat = alpha * x + (1 - alpha) * zold
        for group in groups:
            z[group] = soft_thresholding(x_hat[group] + u[group], lamda / rho)

        # % u-update
        u += (x_hat - z)

        # % diagnostics, reporting, termination checks
        history = (
            objective(A, b, lamda, groups, x, z),  # obj

            np.linalg.norm(x - z),  # r norm
            np.linalg.norm(-rho * (z - zold)),  # s norm

            np.sqrt(n_features) * tol + rtol * max(
                np.linalg.norm(x), np.linalg.norm(-z)),  # eps pri
            np.sqrt(n_features) * tol + rtol * np.linalg.norm(rho * u)  # eps dual
        )

        hist.append(history)
        if history[1] < history[3] and history[2] < history[4]:
            break

    return z, history if return_history else z


def objective(A, b, alpha, groups, x, z):
    # obj = 0
    # for i, group in enumerate(p):
    #     obj = obj + np.linalg.norm(z[group])
    penalty = np.sum([np.linalg.norm(z[g]) for g in groups])
    return .5 * np.sum((A.dot(x) - b) ** 2) + alpha * penalty



"""Validation module for REGAIN."""
import warnings
from functools import partial

import numpy as np
from sklearn.utils.extmath import squared_norm

from regain.norm import l1_norm, node_penalty
from regain.prox import (
    blockwise_soft_thresholding, prox_laplacian, prox_linf, prox_node_penalty,
    soft_thresholding)


def check_norm_prox(function):
    """Validate function and return norm with associated prox."""
    if function == 'laplacian':
        prox = prox_laplacian
        norm = squared_norm
    elif function == 'l1':
        prox = soft_thresholding
        norm = l1_norm
    elif function == 'l2':
        prox = blockwise_soft_thresholding
        norm = np.linalg.norm
    elif function == 'linf':
        prox = prox_linf
        norm = partial(np.linalg.norm, ord=np.inf)
    elif function == 'node':
        prox = prox_node_penalty
        norm = node_penalty
    else:
        raise ValueError("Value of %s not understood." % function)
    return norm, prox, function == 'node'


def check_array_dimensions(
        X, n_dimensions=3, time_on_axis='first', suppress_warn_list=False):
    """Validate input matrix."""
    if isinstance(X, list):
        if not suppress_warn_list:
            warnings.warn(
                "Input data is list; assumed to be a list of matrices "
                "for each time point")
        return X
    if X.ndim == n_dimensions - 1:
        warnings.warn(
            "Input data should have %d"
            " dimensions, found %d. Reshaping input into %s" %
            (n_dimensions, X.ndim, (1, ) + X.shape))
        return X[None, ...]

    elif X.ndim != n_dimensions:
        raise ValueError(
            "Input data should have %d"
            " dimensions, found %d." % (n_dimensions, X.ndim))

    if time_on_axis == 'last':
        return X.transpose(2, 0, 1)  # put time as first dimension

    return X



"""Statistical functions."""
import numpy as np
from sklearn.utils.extmath import squared_norm
from scipy.stats import multivariate_normal


def lognormal_pdf(x, mu, sigma):
    """Lognormal pdf.

    Equivalent to
    stats.lognorm.pdf(x, loc=0, s=sigma, scale=np.exp(mu))

    Reference
    ---------
    https://stackoverflow.com/questions/8870982/how-do-i-get-a-lognormal-distribution-in-python-with-mu-and-sigma
    """
    return np.exp(-0.5 * ((np.log(x) - mu) / sigma) ** 2) \
        / (x * sigma * np.sqrt(2 * np.pi))


def lognormal_logpdf(x, mu, sigma):
    """Lognormal log pdf.

    Equivalent to
    stats.lognorm.logpdf(x, loc=0, s=sigma, scale=np.exp(mu))
    """
    return np.log(lognormal_pdf(x, mu, sigma))


def lognstat(mean, variance):
    """A lognormal distribution with mean m and variance v has mu sigma as."""
    m2 = mean * mean
    mu = np.log(m2 / np.sqrt(variance + m2))
    sigma = np.sqrt(np.log(variance / m2 + 1))
    return mu, sigma


def log_lik_frob(S, D, variance):
    """Frobenius norm log likelihood."""
    logl = -0.5 * (
        S.size * np.log(2. * np.pi * variance) +
        squared_norm(S - D) / variance)
    return logl


def log_likelihood_normal(x, mean, var):
    """Normal log likelihood."""
    logl = -0.5 * (np.log(2 * np.pi * var) + (x - mean)**2 / var)
    # logl2 = stats.norm.logpdf(x, loc=mean, scale=np.sqrt(var))
    # assert logl == logl2, (logl, logl2)
    return logl


def t_mvn_logpdf(X, Cov):
    """Normal log likelihood based on Cov (mu = 0).
    
    Parameters
    ----------
    X : ndarray, shape = (n_samples, n_dimensions, n_times)
        Data tensor.
    Cov : ndarray, shape = (n_dimensions, n_dimensions, n_times)
        Tensor of covariance matrices over time.
    """
    logp = sum(
        x.shape[0] *
        multivariate_normal.logpdf(x, cov=Sigma, allow_singular=True).sum()
        for x, Sigma in zip(X, Cov.T))
    if not isinstance(logp, float):
        logp = sum(logp)
    return logp
